# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from pypowervm.utils import uuid as uuid_utils

import unittest


class TestUUID(unittest.TestCase):
    """Unit tests for the uuid."""

    def test_uuid_conversion(self):
        uuid = '089ffb20-5d19-4a8c-bb80-13650627d985'
        pvm_uuid = uuid_utils.convert_uuid_to_pvm(uuid)
        self.assertEqual(uuid, pvm_uuid)

        uuid = '989ffb20-5d19-4a8c-bb80-13650627d985'
        pvm_uuid = uuid_utils.convert_uuid_to_pvm(uuid)
        self.assertEqual('1' + uuid[1:], pvm_uuid)

        uuid = 'c89ffb20-5d19-4a8c-bb80-13650627d985'
        pvm_uuid = uuid_utils.convert_uuid_to_pvm(uuid)
        self.assertEqual('4' + uuid[1:], pvm_uuid)

    def test_id_or_uuid(self):
        self.assertEqual((False, 123), uuid_utils.id_or_uuid(123))
        self.assertEqual((False, 123), uuid_utils.id_or_uuid('123'))
        self.assertEqual(
            (True, '12345678-abcd-ABCD-0000-0a1B2c3D4e5F'),
            uuid_utils.id_or_uuid('12345678-abcd-ABCD-0000-0a1B2c3D4e5F'))
        self.assertRaises(ValueError, uuid_utils.id_or_uuid,
                          '12345678abcdABCD00000a1B2c3D4e5F')
        # This one has too many digits
        self.assertRaises(ValueError, uuid_utils.id_or_uuid,
                          '12345678-abcd-ABCD-0000-0a1B2c3D4e5F0')



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers/helpers for Performance and Capacity Monitoring (PCM) metrics."""

import abc
import datetime
import pytz
import six

from oslo_log import log as logging

from pypowervm import adapter as adpt
import pypowervm.const as pc
import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap

# Constants that make up the http path
PREFERENCES = 'preferences'
RAW_METRICS = 'RawMetrics'
LONG_TERM_MONITOR = 'LongTermMonitor'
SHORT_TERM_MONITOR = 'ShortTermMonitor'
PCM_SERVICE = 'pcm'

_SYSTEM_NAME = 'SystemName'
_LTM_ENABLED = 'LongTermMonitorEnabled'
_AGG_ENABLED = 'AggregationEnabled'
_STM_ENABLED = 'ShortTermMonitorEnabled'
_COMP_LTM_ENABLED = 'ComputeLTMEnabled'

_UPDATED = 'updated'
_TITLE = 'title'
_PUBLISHED = 'published'
_CATEGORY = 'category'

_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

LOG = logging.getLogger(__name__)


@ewrap.EntryWrapper.pvm_type('ManagedSystemPcmPreference', ns=pc.PCM_NS)
class PcmPref(ewrap.EntryWrapper):
    """Wraps the Performance and Capacity Monitoring preferences."""

    @property
    def system_name(self):
        return self._get_val_str(_SYSTEM_NAME)

    @property
    def ltm_enabled(self):
        """Long Term Monitoring."""
        return self._get_val_bool(_LTM_ENABLED)

    @ltm_enabled.setter
    def ltm_enabled(self, value):
        """Long Term Monitoring."""
        self.set_parm_value(_LTM_ENABLED, u.sanitize_bool_for_api(value))

    @property
    def aggregation_enabled(self):
        """Metrics Aggregation."""
        return self._get_val_bool(_AGG_ENABLED)

    @aggregation_enabled.setter
    def aggregation_enabled(self, value):
        """Metrics Aggregation."""
        self.set_parm_value(_AGG_ENABLED, u.sanitize_bool_for_api(value))

    @property
    def stm_enabled(self):
        """Short Term Monitoring."""
        return self._get_val_bool(_STM_ENABLED)

    @stm_enabled.setter
    def stm_enabled(self, value):
        """Short Term Monitoring.

        Short Term metrics can affect the performance of workloads.  Not
        recommended for production workload.
        """
        self.set_parm_value(_STM_ENABLED, u.sanitize_bool_for_api(value))

    @property
    def compute_ltm_enabled(self):
        """Compute Long Term Monitoring."""
        return self._get_val_bool(_COMP_LTM_ENABLED)

    @compute_ltm_enabled.setter
    def compute_ltm_enabled(self, value):
        """Compute Long Term Monitoring."""
        self.set_parm_value(_COMP_LTM_ENABLED, u.sanitize_bool_for_api(value))


@six.add_metaclass(abc.ABCMeta)
class MonitorMetrics(object):
    """A pseudo wrapper for Monitor metrics.

    The standard pattern of wrapping a response or entry and accessing
    properties for the data can be used, even though this isn't a traditional
    EntryWrapper.
    """
    def __init__(self, entry):
        self.entry = entry

    @staticmethod
    def _str_to_datetime(str_date):
        # The format of the string is one of two ways.
        # Current: 2015-04-30T06:11:35.000-05:00
        # Legacy: 2015-04-30T06:11:35.000Z (the Z was meant to be timezone).
        #
        # The formatter will strip any Z's that may be in the string out.
        str_date = str_date.replace('Z', '-00:00')

        # Separate out the timezone.  Datetime doesn't like formatting time
        # zones, so we pull it out for manual parsing.  It is the 6th digit
        # from the right.
        str_date, str_tz = str_date[:-6], str_date[-6:]

        # We now have the date, without the timezone.
        date = (datetime.datetime.strptime(str_date, _DATETIME_FORMAT).
                replace(tzinfo=pytz.utc))

        # Parse out the timezone.
        tz_oper = str_tz[0]
        tz_hr, tz_min = int(str_tz[1:3]), int(str_tz[4:6])
        tz_delta = datetime.timedelta(hours=tz_hr, minutes=tz_min)

        # Return the date plus/minus the timezone delta.
        return (date + tz_delta) if (tz_oper == '+') else (date - tz_delta)

    @classmethod
    def wrap(cls, response_or_entry):
        if isinstance(response_or_entry, adpt.Response):
            return [cls(entry) for entry in response_or_entry.feed.entries]
        else:
            return cls(response_or_entry)

    @property
    def id(self):
        return self.entry.uuid

    @property
    def published(self):
        return self.entry.properties.get(_PUBLISHED)

    @property
    def published_datetime(self):
        return self._str_to_datetime(self.published)

    @property
    def title(self):
        return self.entry.properties.get(_TITLE)

    @property
    def updated(self):
        return self.entry.properties.get(_UPDATED)

    @property
    def updated_datetime(self):
        return self._str_to_datetime(self.updated)

    @property
    def category(self):
        return self.entry.properties.get(_CATEGORY)

    @property
    def link(self):
        return self.entry.links[None][0]


class LTMMetrics(MonitorMetrics):
    """A pseudo wrapper for Long Term Monitor metrics.

    The standard pattern of wrapping a response or entry and accessing
    properties for the data can be used, even though this isn't a traditional
    EntryWrapper.
    """
    pass


class STMMetrics(MonitorMetrics):
    """A pseudo wrapper for Short Term Monitor metrics.

    The standard pattern of wrapping a response or entry and accessing
    properties for the data can be used, even though this isn't a traditional
    EntryWrapper.
    """
    pass



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pypowervm.tasks import client_storage as clstor
from pypowervm.tests.test_utils import test_wrapper_abc as twrap
from pypowervm.wrappers import virtual_io_server as vios


class TestClientStorage(twrap.TestWrapper):
    file = 'fake_vios_mappings.txt'
    wrapper_class_to_test = vios.VIOS

    def test_udid2scsi(self):
        """Test udid_to_scsi_mapping."""

        maps = self.dwrap.scsi_mappings

        # 2nd mapping has no client adapter
        lpar_id = maps[1].server_adapter.lpar_id
        # Default: ignore orphan
        self.assertIsNone(clstor.udid_to_scsi_mapping(
            self.dwrap, maps[1].backing_storage.udid, lpar_id))
        # Don't ignore orphan
        self.assertEqual(maps[1], clstor.udid_to_scsi_mapping(
            self.dwrap, maps[1].backing_storage.udid, lpar_id,
            ignore_orphan=False))
        # Doesn't work if the LPAR ID is wrong
        self.assertIsNone(clstor.udid_to_scsi_mapping(
            self.dwrap, maps[1].backing_storage.udid, 123,
            ignore_orphan=False))

        # 4th mapping has client adapter but no backing storage
        self.assertIsNone(clstor.udid_to_scsi_mapping(self.dwrap, 'bogus', 22))

    def test_c_wwpn_to_vfc(self):
        """Test c_wwpn_to_vfc_mapping."""
        # Since the first two VFC mappings have no client adapter, this test
        # proves we skip those properly.
        self.assertEqual(self.dwrap.vfc_mappings[5],
                         clstor.c_wwpn_to_vfc_mapping(
                             self.dwrap, 'C05076065A7C02E3'))

        # This works with (limited) craziness in the WWPN format
        self.assertEqual(self.dwrap.vfc_mappings[5],
                         clstor.c_wwpn_to_vfc_mapping(
                             self.dwrap, 'c0:50:76:06:5a:7c:02:e3'))

        # Not found
        self.assertIsNone(clstor.c_wwpn_to_vfc_mapping(
            self.dwrap, 'ab:cd:ef:01:23:45:67:89'))



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""LPAR, the EntryWrapper for LogicalPartition."""

from oslo_log import log as logging

import pypowervm.const as pc
from pypowervm.i18n import _
import pypowervm.util as u
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.entry_wrapper as ewrap

LOG = logging.getLogger(__name__)

_LPAR_MIG_STG_VIOS_DATA_STATUS = 'MigrationStorageViosDataStatus'
_LPAR_MIG_STG_VIOS_DATA_TIME = 'MigrationStorageViosDataTimestamp'
_LPAR_RR = 'RemoteRestartCapable'
_LPAR_SRR = 'SimplifiedRemoteRestartCapable'
_LPAR_HAS_DED_PROCS_FOR_MIG = 'HasDedicatedProcessorsForMigration'
_LPAR_SUSPEND_CAP = 'SuspendCapable'
_LPAR_MIG_DISABLE = 'MigrationDisable'
_LPAR_MIG_STATE = 'MigrationState'
_LPAR_RR_STATE = 'RemoteRestartState'
_LPAR_PRI_PGING_SVC_PART = 'PrimaryPagingServicePartition'
_LPAR_POWER_MGT_MODE = 'PowerManagementMode'
_LPAR_SEC_PGING_SVC_PART = 'SecondaryPagingServicePartition'
_LPAR_USES_HSL_OPTICONN = 'UsesHighSpeedLinkOpticonnect'
_LPAR_USES_VIRT_OPTICONN = 'UsesVirtualOpticonnect'
_LPAR_VFC_CLIENT_ADPTS = 'VirtualFibreChannelClientAdapters'
_LPAR_VSCSI_CLIENT_ADPTS = 'VirtualSCSIClientAdapters'
_LPAR_RESTRICTED_IO = 'IsRestrictedIOPartition'
_LPAR_STG_DEV_UDID = 'StorageDeviceUniqueDeviceID'
_LPAR_DES_IPL_SRC = 'DesignatedIPLSource'
_LPAR_DED_VNICS = 'DedicatedVirtualNICs'
_LPAR_BOOTLIST_INFO = 'BootListInformation'

_LPAR_EL_ORDER = bp.BP_EL_ORDER + (
    _LPAR_MIG_STG_VIOS_DATA_STATUS, _LPAR_MIG_STG_VIOS_DATA_TIME, _LPAR_RR,
    _LPAR_SRR, _LPAR_HAS_DED_PROCS_FOR_MIG, _LPAR_SUSPEND_CAP,
    _LPAR_MIG_DISABLE, _LPAR_MIG_STATE, _LPAR_RR_STATE,
    _LPAR_PRI_PGING_SVC_PART, _LPAR_POWER_MGT_MODE, _LPAR_SEC_PGING_SVC_PART,
    _LPAR_USES_HSL_OPTICONN, _LPAR_USES_VIRT_OPTICONN, _LPAR_VFC_CLIENT_ADPTS,
    _LPAR_VSCSI_CLIENT_ADPTS, _LPAR_RESTRICTED_IO, _LPAR_STG_DEV_UDID,
    _LPAR_DES_IPL_SRC, _LPAR_DED_VNICS, _LPAR_BOOTLIST_INFO)


class IPLSrc(object):
    """Mirror of IPLSource.Enum (relevant to IBMi partitions only).

    Valid values for:
    - LPAR.desig_ipl_src
    - 'iIPLsource' param in pypowervm.power.power_on.

    Example usage:
    - ilpar.desig_ipl_src = IPLSrc.C
      ilpar.update()
    - power_on(..., add_parms={IPLSrc.KEY: IPLSrc.A, ...})
    """
    KEY = 'iIPLsource'
    A = 'a'
    B = 'b'
    C = 'c'
    D = 'd'
    UNKNOWN = 'Unknown'
    ALL_VALUES = (A, B, C, D, UNKNOWN)


class RRState(object):
    """Remote Restart states - mirror of PartitionRemoteRestart.Enum."""
    INVALID = "Invalid"
    RR_ABLE = "Remote_Restartable"
    SRC_RRING = "Source_Remote_Restarting"
    DEST_RRING = "Destination_Remote_Restarting"
    REM_RESTARTED = "Remote_Restarted"
    PROF_RESTORED = "Profile_Restored"
    RES_STG_DEV_UPD_FAIL = "Reserved_Storage_Device_Update_Failed"
    FORCED_SRC_RESTART = "Forced_Source_Side_Restart"
    SRC_CLEANUP_FAIL = "Source_Side_Cleanup_Failed"
    RES_STG_DEV_UPD_FAIL_W_OVRD = ("Reserved_Storage_Device_Update_Failed_With"
                                   "_Override")
    RR_ABLE_SUSPENDED = "Remote_Restartable_Suspended"
    LOC_UPD_FAIL = "Local_Update_Failed"
    PART_UPD = "Partial_Update"
    STALE_DATA = "Stale_Data"
    LOC_DATA_VALID = "Local_Data_Valid"
    OUT_OF_SPACE = "Out_Of_Space"
    LOC_DATA_INVALID = "Local_Data_Invalid"
    DEST_RR_ED = "Destination_Remote_Restarted"
    SRC_RRING_SUSPENDED = "Source_Remote_Restarting_Suspended"
    LOC_STG_UPD_FAIL = "Local_Storage_Update_Failed"
    PG_DEV_UPD_OVRD = "Page_Device_Update_Override"


class BootStorageType(object):
    """Enumeration of possible storage connection methods for devices."""
    VSCSI = 'vscsi'
    VFC = 'npiv'
    UNKNOWN = 'Unknown'
    ALL_VALUES = (VSCSI, VFC, UNKNOWN)


@ewrap.EntryWrapper.pvm_type('LogicalPartition',
                             child_order=_LPAR_EL_ORDER)
class LPAR(bp.BasePartition, ewrap.WrapperSetUUIDMixin):

    @classmethod
    def bld(cls, adapter, name, mem_cfg, proc_cfg, env=bp.LPARType.AIXLINUX,
            io_cfg=None):
        """Creates an LPAR wrapper.

        Thin wrapper around BasePartition._bld_base, defaulting env.
        """
        return super(LPAR, cls)._bld_base(adapter, name, mem_cfg, proc_cfg,
                                          env, io_cfg)

    def _can_modify(self, dlpar_cap, cap_desc):
        """Checks to determine if the LPAR can be modified.

        :param dlpar_cap: The appropriate DLPAR attribute to validate.  Only
                          used if system is active.
        :param cap_desc: A translated string indicating the DLPAR capability.
        :return capable: True if HW can be added/removed.  False otherwise.
        :return reason: A translated message that will indicate why it was not
                        capable of modification.  If capable is True, the
                        reason will be None.
        """
        # If we are in the LPAR, we have access to the operating system type.
        # If it is an OS400 type, then we can add/remove HW no matter what.
        if self.env == bp.LPARType.OS400:
            return True, None
        return super(LPAR, self)._can_modify(dlpar_cap, cap_desc)

    def can_lpm(self, host_w, migr_data=None):
        """Determines if a LPAR is ready for Live Partition Migration.

        This check validates that the target system is capable of
        handling the LPAR if the LPAR is an IBMi.  It simply validates that
        the LPAR has the essential capabilities in place for a LPM operation.

        :param host_w: The host wrapper for the system.
        :param migr_data: The dictionary of migration data for the target host.
                          If parameters are not passed in, will skip the check
                          and let the low levels surface related error.
                          The supported key today is:
                          - ibmi_lpar_mobility_capable: Boolean
                          TODO(IBM): add more destination checks here. Ex.
                          migrate an AIX or IBMi VM to a Linux only host.
        :return capable: True if the LPAR is LPM capable.  False otherwise.
        :return reason: A translated message that will indicate why it was not
                        capable of LPM.  If capable is True, the reason will
                        be None.
        """
        # First check is the not activated state
        if self.state != bp.LPARState.RUNNING:
            return False, _("LPAR is not in an active state.")

        if self.env == bp.LPARType.OS400:
            # IBM i does not require RMC, but does need to check for target
            # host and source host are capable for IBMi mobility and
            # restricted I/O.

            if migr_data is not None:
                c = migr_data.get('ibmi_lpar_mobility_capable')
                if c is not None and not c:
                    return False, _('Target system does not have the IBM i'
                                    ' LPAR Mobility Capability.')

            if not self.restrictedio:
                return False, _('IBM i LPAR does not have restricted I/O.')

            c = host_w.get_capabilities().get('ibmi_lpar_mobility_capable')
            if not c:
                return False, _('Source system does not have the IBM i'
                                ' LPAR Mobility Capability.')

        elif self.rmc_state != bp.RMCState.ACTIVE:
            return False, _('LPAR does not have an active RMC connection.')

        if self.is_mgmt_partition:
            return False, _('LPAR is the management partition')

        c = self.capabilities
        if not (c.mem_dlpar and c.proc_dlpar):
            return False, _('LPAR is not available for LPM due to missing '
                            'DLPAR capabilities.')
        return True, None

    @property
    def migration_state(self):
        """See PartitionMigrationStateEnum.

        e.g. 'Not_Migrating', 'Migration_Starting', 'Migration_Failed', etc.
        Defaults to 'Not_Migrating'
        """
        return self._get_val_str(_LPAR_MIG_STATE, 'Not_Migrating')

    @property
    def rr_enabled(self):
        """Deprecated (n/a for NovaLink) - use srr_enabled instead."""
        import warnings
        warnings.warn(_("This is not the property you are looking for.  Use "
                        "srr_enabled in a NovaLink environment."),
                      DeprecationWarning)
        return None

    @rr_enabled.setter
    def rr_enabled(self, value):
        """Deprecated (n/a for NovaLink) - use srr_enabled instead."""
        import warnings
        warnings.warn(_("This is not the property you are looking for.  Use "
                        "srr_enabled in a NovaLink environment."),
                      DeprecationWarning)

    @property
    def rr_state(self):
        """Deprecated (n/a for NovaLink) - use srr_enabled instead."""
        import warnings
        warnings.warn(_("This is not the property you are looking for.  Use "
                        "srr_enabled in a NovaLink environment."),
                      DeprecationWarning)
        return None

    @property
    def srr_enabled(self):
        """Simplied remote restart.

        :returns: Returns SRR config boolean
        """
        return self._get_val_bool(_LPAR_SRR, False)

    @srr_enabled.setter
    def srr_enabled(self, value):
        self.set_parm_value(_LPAR_SRR, u.sanitize_bool_for_api(value),
                            attrib=pc.ATTR_KSV120)

    @property
    def restrictedio(self):
        return self._get_val_bool(_LPAR_RESTRICTED_IO, False)

    @restrictedio.setter
    def restrictedio(self, value):
        self.set_parm_value(_LPAR_RESTRICTED_IO,
                            u.sanitize_bool_for_api(value))

    @property
    def desig_ipl_src(self):
        """Designated IPL Source - see IPLSrc enumeration."""
        return self._get_val_str(_LPAR_DES_IPL_SRC)

    @desig_ipl_src.setter
    def desig_ipl_src(self, value):
        """Designated IPL Source - see IPLSrc enumeration."""
        if value not in IPLSrc.ALL_VALUES:
            raise ValueError(_("Invalid IPLSrc '%s'.") % value)
        self.set_parm_value(_LPAR_DES_IPL_SRC, value)

    def set_uuid(self, value):
        # LPAR uuids must be uppercase.
        up_uuid = str(value).upper()
        super(LPAR, self).set_uuid(up_uuid)
        self.set_parm_value(bp._BP_UUID, up_uuid)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
import testtools

from pypowervm.tasks import scsi_mapper
from pypowervm.tests.tasks import util as tju
from pypowervm.tests import test_fixtures as fx
from pypowervm.wrappers import storage as pvm_stor
from pypowervm.wrappers import virtual_io_server as pvm_vios

VIO_MULTI_MAP_FILE = 'vio_multi_vscsi_mapping.txt'
VIO_MULTI_MAP_FILE2 = 'fake_vios_mappings.txt'
LPAR_UUID = '42AD4FD4-DC64-4935-9E29-9B7C6F35AFCC'


class TestSCSIMapper(testtools.TestCase):

    def setUp(self):
        super(TestSCSIMapper, self).setUp()
        # Common Adapter
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        # Don't really sleep
        self.useFixture(fx.SleepFx())

        # Fake URI
        mock_crt_href_p = mock.patch('pypowervm.wrappers.virtual_io_server.'
                                     'VSCSIMapping.crt_related_href')
        self.mock_crt_href = mock_crt_href_p.start()
        self.addCleanup(mock_crt_href_p.stop)
        href = ('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                'c5d782c7-44e4-3086-ad15-b16fb039d63b/LogicalPartition/' +
                LPAR_UUID)
        self.mock_crt_href.return_value = href

        # Mock the delay function, by overriding the sleep
        mock_delay_p = mock.patch('time.sleep')
        self.mock_delay = mock_delay_p.start()
        self.addCleanup(mock_delay_p.stop)

        self.v1resp = tju.load_file(VIO_MULTI_MAP_FILE, self.adpt)
        self.v1wrap = pvm_vios.VIOS.wrap(self.v1resp)
        self.v2resp = tju.load_file(VIO_MULTI_MAP_FILE2, self.adpt)
        self.v2wrap = pvm_vios.VIOS.wrap(self.v2resp)

    def test_mapping(self):
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was added to existing
        def validate_update(*kargs, **kwargs):
            vios_w = kargs[0]
            self.assertEqual(6, len(vios_w.scsi_mappings))
            self.assertEqual(vios_w.scsi_mappings[0].client_adapter,
                             vios_w.scsi_mappings[4].client_adapter)
            self.assertEqual(vios_w.scsi_mappings[0].server_adapter,
                             vios_w.scsi_mappings[4].server_adapter)
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Create the new storage dev
        pv = pvm_stor.PV.bld(self.adpt, 'pv_name', 'pv_udid')

        # Run the code
        scsi_mapper.add_vscsi_mapping('host_uuid', 'vios_uuid', LPAR_UUID,
                                      pv)

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        # And the VIOS was "looked up"
        self.assertEqual(1, self.adpt.read.call_count)

        # Now do it again, but passing the vios wrapper
        self.adpt.update_by_path.reset_mock()
        self.adpt.read.reset_mock()
        scsi_mapper.add_vscsi_mapping('host_uuid', self.v1wrap, LPAR_UUID, pv)
        # Since the mapping already existed, our update mock was not called
        self.assertEqual(0, self.adpt.update_by_path.call_count)
        # And the VIOS was not "looked up"
        self.assertEqual(0, self.adpt.read.call_count)

    def test_mapping_retry(self):
        """Tests that a mapping function will be retried."""
        # Mock Data.  Need to load this once per retry, or else the mappings
        # get appended with each other.
        self.adpt.read.side_effect = [
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt),
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt),
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt)]

        global attempt_count
        attempt_count = 0

        # Validate that the mapping was added to existing.  First few times
        # through loop, force a retry exception
        def validate_update(*kargs, **kwargs):
            global attempt_count
            attempt_count += 1

            if attempt_count == 3:
                vios_w = kargs[0]
                self.assertEqual(6, len(vios_w.scsi_mappings))
                return vios_w.entry
            else:
                tju.raiseRetryException()

        self.adpt.update_by_path.side_effect = validate_update

        # Create the new storage dev
        pv = pvm_stor.PV.bld(self.adpt, 'pv_name', 'pv_udid')

        # Run the code
        scsi_mapper.add_vscsi_mapping('host_uuid', 'vios_uuid', LPAR_UUID,
                                      pv)

        # Make sure that our validation code above was invoked
        self.assertEqual(3, self.adpt.update_by_path.call_count)
        self.assertEqual(3, attempt_count)

    def test_mapping_new_mapping(self):
        """Fuse limit, slot number, LUA via add_vscsi_mapping."""
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was added to existing
        def validate_update(*args, **kwargs):
            vios_w = args[0]
            self.assertEqual(6, len(vios_w.scsi_mappings))

            new_map = vios_w.scsi_mappings[5]
            # Make sure that the adapters do not match
            self.assertNotEqual(vios_w.scsi_mappings[0].client_adapter,
                                new_map.client_adapter)
            self.assertNotEqual(vios_w.scsi_mappings[0].server_adapter,
                                new_map.server_adapter)
            # Make sure we got the right slot number and LUA
            self.assertEqual(23, new_map.client_adapter.lpar_slot_num)
            self.assertEqual('the_lua', new_map.target_dev.lua)
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Create the new storage dev
        pv = pvm_stor.PV.bld(self.adpt, 'pv_name', 'pv_udid')

        # Run the code
        # While we're here, make sure lpar_slot_num and lua go through.  This
        # validates those kwargs in build_vscsi_mapping too.
        scsi_mapper.add_vscsi_mapping(
            'host_uuid', 'vios_uuid', LPAR_UUID, pv, fuse_limit=5,
            lpar_slot_num=23, lua='the_lua')

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)

    def test_add_map(self):
        """Tests the add_map method."""
        pv = pvm_stor.PV.bld(self.adpt, 'pv_name', 'pv_udid')

        scsi_map = scsi_mapper.build_vscsi_mapping('host_uuid', self.v1wrap,
                                                   LPAR_UUID, pv,
                                                   lpar_slot_num=23)

        # Get the original count
        orig_mappings = len(self.v1wrap.scsi_mappings)

        # Add the actual mapping
        resp1 = scsi_mapper.add_map(self.v1wrap, scsi_map)
        self.assertIsNotNone(resp1)
        self.assertIsInstance(resp1, pvm_vios.VSCSIMapping)

        # Assert that the desired client slot number was set
        self.assertEqual(resp1.client_adapter.lpar_slot_num, 23)

        # The mapping should return as None, as it is already there.
        resp2 = scsi_mapper.add_map(self.v1wrap, scsi_map)
        self.assertIsNone(resp2)

        # Make sure only one was added.
        self.assertEqual(orig_mappings + 1, len(self.v1wrap.scsi_mappings))

        # Now make sure the mapping added can be found
        found = scsi_mapper.find_maps(self.v1wrap.scsi_mappings, LPAR_UUID,
                                      stg_elem=pv)
        self.assertEqual(1, len(found))
        self.assertEqual(scsi_map, found[0])

    def test_remove_storage_vopt(self):
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was removed from existing
        def validate_update(*kargs, **kwargs):
            vios_w = kargs[0]
            self.assertEqual(4, len(vios_w.scsi_mappings))
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Run the code
        media_name = 'bldr1_dfe05349_kyleh_config.iso'
        vios, remel = scsi_mapper.remove_vopt_mapping(
            self.adpt, 'fake_vios_uuid', 2, media_name=media_name)

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.VOptMedia)
        # And the VIOS was "looked up"
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(self.v1resp.atom, vios.entry)

        # Now do it again, but passing the vios wrapper and the client UUID.
        # Match by UDID this time.
        media_udid = '0ebldr1_dfe05349_kyleh_config.iso'
        vios_wrap = pvm_vios.VIOS.wrap(
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt))
        self.adpt.update_by_path.reset_mock()
        self.adpt.read.reset_mock()
        vios, remel = scsi_mapper.remove_vopt_mapping(
            self.adpt, vios_wrap, LPAR_UUID, udid=media_udid)
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.VOptMedia)
        # But the VIOS was not "looked up"
        self.assertEqual(0, self.adpt.read.call_count)
        self.assertEqual(vios_wrap.entry, vios.entry)

    def test_remove_storage_vopt_no_name_specified(self):
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was removed from existing
        def validate_update(*kargs, **kwargs):
            vios_w = kargs[0]
            self.assertEqual(4, len(vios_w.scsi_mappings))
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Run the code
        vios, remel = scsi_mapper.remove_vopt_mapping(
            self.adpt, 'fake_vios_uuid', 2, media_name=None)

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.VOptMedia)
        self.assertEqual(self.v1resp.atom, vios.entry)

    def test_remove_storage_vopt_retry(self):
        """Tests removing the storage vOpt with multiple retries."""
        # Mock Data.  The retry will call this three times.  They have to
        # be indepdent loads, otherwise the data gets re-used and the remove
        # will not be properly invoked.
        self.adpt.read.side_effect = [
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt),
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt),
            tju.load_file(VIO_MULTI_MAP_FILE, self.adpt)]

        global attempt_count
        attempt_count = 0

        # Validate that the mapping was removed from existing.  First few
        # loops, force a retry
        def validate_update(*kargs, **kwargs):
            global attempt_count
            attempt_count += 1

            if attempt_count == 3:
                vios_w = kargs[0]
                self.assertEqual(4, len(vios_w.scsi_mappings))
                return vios_w.entry
            else:
                tju.raiseRetryException()

        self.adpt.update_by_path.side_effect = validate_update

        # Run the code
        media_name = 'bldr1_dfe05349_kyleh_config.iso'
        remel = scsi_mapper.remove_vopt_mapping(
            self.adpt, 'fake_vios_uuid', 2, media_name=media_name)[1]

        # Make sure that our validation code above was invoked
        self.assertEqual(3, self.adpt.update_by_path.call_count)
        self.assertEqual(3, attempt_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.VOptMedia)

    def _test_remove_storage_vdisk(self, *args, **kwargs):
        """Helper to test remove_storage_vdisk with various arguments."""
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was removed from existing
        def validate_update(*kargs, **kwa):
            vios_w = kargs[0]
            self.assertEqual(4, len(vios_w.scsi_mappings))
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Run the code
        vios, remel = scsi_mapper.remove_vdisk_mapping(*args, **kwargs)

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.VDisk)
        self.assertEqual(self.v1resp.atom, vios.entry)

    def test_remove_storage_vdisk_name(self):
        self._test_remove_storage_vdisk(
            self.adpt, 'fake_vios_uuid', 2, disk_names=['Ubuntu1410'])

    def test_remove_storage_vdisk_udid(self):
        self._test_remove_storage_vdisk(
            self.adpt, 'fake_vios_uuid', 2,
            udids=['0300025d4a00007a000000014b36d9deaf.1'])

    def _test_remove_storage_lu(self, *args, **kwargs):
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was removed from existing
        def validate_update(*kargs, **kwa):
            vios_w = kargs[0]
            self.assertEqual(4, len(vios_w.scsi_mappings))
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Run the code
        vios, remel = scsi_mapper.remove_lu_mapping(*args, **kwargs)

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.LU)
        self.assertEqual(self.v1resp.atom, vios.entry)

    def test_remove_storage_lu_all(self):
        self._test_remove_storage_lu(self.adpt, 'fake_vios_uuid', 2)

    def test_remove_storage_lu_udid(self):
        self._test_remove_storage_lu(
            self.adpt, 'fake_vios_uuid', 2,
            udids=['270c88f8e2d36711e490ce40f2e95daf30a6d61c0dee5ec6f6a011b300'
                   'b9d0830d'])

    def _test_remove_pv_mapping(self, *args, **kwargs):
        # Mock Data
        self.adpt.read.return_value = self.v1resp

        # Validate that the mapping was removed to existing
        def validate_update(*kargs, **kwa):
            vios_w = kargs[0]
            self.assertEqual(4, len(vios_w.scsi_mappings))
            return vios_w.entry

        self.adpt.update_by_path.side_effect = validate_update

        # Run the code
        vios, remel = scsi_mapper.remove_pv_mapping(*args, **kwargs)

        # Make sure that our validation code above was invoked
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, len(remel))
        self.assertIsInstance(remel[0], pvm_stor.PV)
        self.assertEqual(self.v1resp.atom, vios.entry)

    def test_remove_pv_mapping_name(self):
        self._test_remove_pv_mapping(self.adpt, 'fake_vios_uuid', 2, 'hdisk10')

    def test_remove_pv_mapping_udid(self):
        self._test_remove_pv_mapping(
            self.adpt, 'fake_vios_uuid', 2, None,
            udid='01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEODgwMDAwMDAwMDAwMDA2MA'
                 '==')

    def test_detach_storage(self):
        """Detach storage from some mappings."""
        # In v1wrap, all five maps are associated with LPAR 2
        num_matches = 5
        self.assertEqual(num_matches, len(self.v1wrap.scsi_mappings))
        # Beforehand, four of them have storage and one does not.
        self.assertEqual(4, len(
            [sm.backing_storage for sm in self.v1wrap.scsi_mappings
             if sm.backing_storage is not None]))
        removals = scsi_mapper.detach_storage(self.v1wrap, 2)
        # The number of mappings is the same afterwards.
        self.assertEqual(num_matches, len(self.v1wrap.scsi_mappings))
        # But now the mappings have no storage
        for smap in self.v1wrap.scsi_mappings:
            self.assertIsNone(smap.backing_storage)
        # The return list contains all the mappings
        self.assertEqual(num_matches, len(removals))
        # The return list members contain the storage (the four that had it
        # beforehand).
        self.assertEqual(4, len([sm.backing_storage for sm in removals
                                 if sm .backing_storage is not None]))

        # In v2wrap, there are four VOptMedia mappings
        num_matches = 4
        match_class = pvm_stor.VOptMedia
        # Number of mappings should be the same before and after.
        len_before = len(self.v2wrap.scsi_mappings)
        self.assertEqual(
            num_matches, len([1 for sm in self.v2wrap.scsi_mappings
                              if isinstance(sm.backing_storage, match_class)]))
        removals = scsi_mapper.detach_storage(
            self.v2wrap, None, match_func=scsi_mapper.gen_match_func(
                match_class))
        self.assertEqual(num_matches, len(removals))
        # The number of mappings is the same as beforehand.
        self.assertEqual(len_before, len(self.v2wrap.scsi_mappings))
        # Now there should be no mappings with VOptMedia
        self.assertEqual(
            0, len([1 for sm in self.v2wrap.scsi_mappings
                    if isinstance(sm.backing_storage, match_class)]))
        # The removals contain the storage
        self.assertEqual(
            num_matches, len([1 for sm in removals
                              if isinstance(sm.backing_storage, match_class)]))

    def test_find_maps(self):
        """find_maps() tests not covered elsewhere."""
        maps = self.v1wrap.scsi_mappings
        # Specifying both match_func and stg_elem raises ValueError
        self.assertRaises(ValueError, scsi_mapper.find_maps, maps, 1,
                          match_func=isinstance, stg_elem='foo')
        # Omitting match_func and stg_elem matches all entries with specified
        # LPAR ID.
        # For LPAR ID 2, that should be all of 'em.
        matches = scsi_mapper.find_maps(maps, 2)
        self.assertEqual(len(maps), len(matches))
        for exp, act in zip(maps, matches):
            self.assertEqual(exp, act)
        # For the right LPAR UUID, that should be all of 'em.
        matches = scsi_mapper.find_maps(maps, LPAR_UUID)
        self.assertEqual(len(maps), len(matches))
        for exp, act in zip(maps, matches):
            self.assertEqual(exp, act)
        # For the wrong LPAR ID, it should be none of 'em.
        matches = scsi_mapper.find_maps(maps, 1)
        self.assertEqual(0, len(matches))
        # For the wrong LPAR UUID, it should be none of 'em.
        matches = scsi_mapper.find_maps(maps, LPAR_UUID[:35] + '0')
        self.assertEqual(0, len(matches))
        # Specific storage element generates match func for that element.
        matches = scsi_mapper.find_maps(maps, 2,
                                        stg_elem=maps[2].backing_storage)
        self.assertEqual(1, len(matches))
        self.assertEqual(maps[2], matches[0])
        # Test find maps when client lpar id is not specified and backing
        # storage is given
        matches = scsi_mapper.find_maps(maps, None,
                                        stg_elem=maps[2].backing_storage)
        self.assertEqual(1, len(matches))
        self.assertEqual(maps[2], matches[0])

        # All the mappings in VIO_MULTI_MAP_FILE are "complete".  Now play with
        # some that aren't.
        maps = self.v2wrap.scsi_mappings
        # Map 0 has only a server adapter.  We should find it if we specify the
        # LPAR ID...
        matches = scsi_mapper.find_maps(maps, 27, include_orphans=True)
        self.assertEqual(maps[0], matches[0])
        # ...but only if allowing orphans
        matches = scsi_mapper.find_maps(maps, 27, include_orphans=False)
        self.assertEqual(0, len(matches))
        # Matching by LPAR UUID.  Maps 12, 25, and 26 have this UUID...
        uuid = '0C0A6EBE-7BF4-4707-8780-A140F349E42E'
        matches = scsi_mapper.find_maps(maps, uuid, include_orphans=True)
        self.assertEqual(3, len(matches))
        self.assertEqual(maps[12], matches[0])
        self.assertEqual(maps[25], matches[1])
        self.assertEqual(maps[26], matches[2])
        # ...but 25 is an orphan (no client adapter).
        uuid = '0C0A6EBE-7BF4-4707-8780-A140F349E42E'
        matches = scsi_mapper.find_maps(maps, uuid)
        self.assertEqual(2, len(matches))
        self.assertEqual(maps[12], matches[0])
        self.assertEqual(maps[26], matches[1])

    def test_separate_mappings(self):
        client_href = ('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                       '726e9cb3-6576-3df5-ab60-40893d51d074/LogicalPartition/'
                       '0C0A6EBE-7BF4-4707-8780-A140F349E42E')
        sep = scsi_mapper._separate_mappings(self.v2wrap, client_href)
        self.assertEqual(2, len(sep))
        self.assertEqual(
            {'1eU8246.L2C.0604C7A-V1-C13', '1eU8246.L2C.0604C7A-V1-C25'},
            set(sep.keys()))
        self.assertEqual(sep['1eU8246.L2C.0604C7A-V1-C13'][0],
                         self.v2wrap.scsi_mappings[-1])

    def test_index_mappings(self):
        idx = scsi_mapper.index_mappings(self.v2wrap.scsi_mappings)

        self.assertEqual({
            'by-lpar-id', 'by-lpar-uuid', 'by-storage-udid'}, set(idx.keys()))

        exp_lpar_ids = ('2', '5', '6', '7', '10', '11', '12', '13', '14', '15',
                        '16', '17', '18', '19', '20', '21', '22', '23', '24',
                        '27', '28', '29', '33', '35', '36', '39', '40')
        self.assertEqual(set(exp_lpar_ids), set(idx['by-lpar-id'].keys()))
        # Each mapping has a different LPAR ID, so each LPAR ID only has one
        # mapping
        for lpar_id in exp_lpar_ids:
            maplist = idx['by-lpar-id'][lpar_id]
            self.assertEqual(1, len(maplist))
            self.assertIsInstance(maplist[0], pvm_vios.VSCSIMapping)
            self.assertEqual(lpar_id, str(maplist[0].server_adapter.lpar_id))

        # Not all mappings have client_lpar_href, so this list is shorter.
        exp_lpar_uuids = ('0C0A6EBE-7BF4-4707-8780-A140F349E42E',
                          '0FB69DD7-4B93-4C09-8916-8BC9821ABAAC',
                          '263EE77B-AD6E-4920-981A-4B7D245B8571',
                          '292ACAF5-C96B-447A-8C7E-7503D80AA33E',
                          '32AA6AA5-CCE6-4523-860C-0852455036BE',
                          '3CE30EC6-C98A-4A58-A764-09DAC7C324BC',
                          '615C9134-243D-4A11-93EB-C0556664B761',
                          '7CFDD55B-E0D7-4B8C-8254-9305E31BB1DC')
        self.assertEqual(set(exp_lpar_uuids), set(idx['by-lpar-uuid'].keys()))
        # Of ten mappings with client_lpar_href, three have the same UUID.
        for lpar_uuid in exp_lpar_uuids:
            maplist = idx['by-lpar-uuid'][lpar_uuid]
            for smap in maplist:
                self.assertIsInstance(smap, pvm_vios.VSCSIMapping)
                self.assertTrue(smap.client_lpar_href.endswith(lpar_uuid))
            if lpar_uuid == '0C0A6EBE-7BF4-4707-8780-A140F349E42E':
                self.assertEqual(3, len(maplist))
            else:
                self.assertEqual(1, len(maplist))

        # Only five mappings have storage, and all are different
        self.assertEqual(5, len(idx['by-storage-udid'].keys()))
        for sudid in idx['by-storage-udid']:
            self.assertEqual(1, len(idx['by-storage-udid'][sudid]))

    def test_gen_match_func(self):
        """Tests for gen_match_func."""

        # Class must match
        mfunc = scsi_mapper.gen_match_func(str)
        self.assertFalse(mfunc(1))
        self.assertTrue(mfunc('foo'))

        # Match names
        elem = mock.Mock()
        elem.name = 'foo'
        # 'False' names/prefixes ignored
        mfunc = scsi_mapper.gen_match_func(mock.Mock, names=[])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes=[])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, names=[], prefixes=[])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, names=['bar', 'baz'])
        self.assertFalse(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, names=['bar', 'foobar',
                                                             'baz'])
        self.assertFalse(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock,
                                           names=['bar', 'foo', 'baz'])
        self.assertTrue(mfunc(elem))

        # Prefixes are ignored if names specified
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes='x',
                                           names=['bar', 'foo', 'baz'])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, names=['bar', 'baz'],
                                           prefixes=['f'])
        self.assertFalse(mfunc(elem))

        # Prefixes
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes=['f'])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes=['foo'])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes=['foo', 'x'])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes=['x'])
        self.assertFalse(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, prefixes=['xfoo', 'foox',
                                                                'xfoox'])
        self.assertFalse(mfunc(elem))

        # Alternate key for the name property
        elem = mock.Mock(alt_name='foo')
        mfunc = scsi_mapper.gen_match_func(mock.Mock, name_prop='alt_name',
                                           names=[])
        self.assertTrue(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, names=['bar', 'baz'])
        self.assertFalse(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, name_prop='alt_name',
                                           names=['bar', 'baz'])
        self.assertFalse(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock,
                                           names=['bar', 'foo', 'baz'])
        self.assertFalse(mfunc(elem))
        mfunc = scsi_mapper.gen_match_func(mock.Mock, name_prop='alt_name',
                                           names=['bar', 'foo', 'baz'])
        self.assertTrue(mfunc(elem))



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from pypowervm import exceptions as pvm_ex
from pypowervm.tasks import vopt
from pypowervm.tests import test_fixtures as pvm_fx


class TestVOpt(testtools.TestCase):
    """Tests the vopt file."""

    def setUp(self):
        super(TestVOpt, self).setUp()

        self.apt = self.useFixture(pvm_fx.AdapterFx()).adpt

        # Wipe out the static variables, so that the re-validate is called
        vopt._cur_vios_uuid = None
        vopt._cur_vg_uuid = None

    @mock.patch('pypowervm.wrappers.storage.VG.get')
    @mock.patch('pypowervm.tasks.partition.get_active_vioses')
    def test_validate_vopt_vg1(self, mock_vios_get, mock_vg_get):
        """One VIOS, rootvg found; locals are set."""
        # Init objects to test with
        mock_vg = mock.Mock()
        mock_vg.configure_mock(name='rootvg',
                               uuid='1e46bbfd-73b6-3c2a-aeab-a1d3f065e92f',
                               vmedia_repos=['repo'])
        mock_vg_get.return_value = [mock_vg]
        mock_vios = mock.Mock()
        mock_vios.configure_mock(name='the_vios', uuid='vios_uuid',
                                 rmc_state='active')
        mock_vios_get.return_value = [mock_vios]

        # Run
        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(self.apt)
        self.assertEqual('vios_uuid', vio_ret_uuid)
        self.assertEqual('1e46bbfd-73b6-3c2a-aeab-a1d3f065e92f', vg_ret_uuid)

        # Validate
        self.assertEqual('1e46bbfd-73b6-3c2a-aeab-a1d3f065e92f',
                         vopt._cur_vg_uuid)
        self.assertEqual('vios_uuid', vopt._cur_vios_uuid)

    @mock.patch('pypowervm.tasks.partition.get_active_vioses')
    @mock.patch('pypowervm.wrappers.storage.VG.get')
    @mock.patch('pypowervm.wrappers.storage.VMediaRepos.bld')
    def test_validate_vopt_vg2(self, mock_vmr_bld, mock_vg_get, mock_vios_get):
        """Dual VIOS, multiple VGs, repos on non-rootvg."""
        vwrap1 = mock.Mock()
        vwrap1.configure_mock(name='vio1', rmc_state='active', uuid='vio_id1',
                              is_mgmt_partition=False)
        vwrap2 = mock.Mock()
        vwrap2.configure_mock(name='vio2', rmc_state='active', uuid='vio_id2',
                              is_mgmt_partition=False)
        mock_vios_get.return_value = [vwrap1, vwrap2]
        vg1 = mock.Mock()
        vg1.configure_mock(name='rootvg', vmedia_repos=[], uuid='vg1')
        vg2 = mock.Mock()
        vg2.configure_mock(name='other1vg', vmedia_repos=[], uuid='vg2')
        vg3 = mock.Mock()
        vg3.configure_mock(name='rootvg', vmedia_repos=[], uuid='vg3')
        vg4 = mock.Mock()
        vg4.configure_mock(name='other2vg', vmedia_repos=[1], uuid='vg4')

        # 1: Find the media repos on non-rootvg on the second VIOS
        mock_vg_get.side_effect = [[vg1, vg2], [vg3, vg4]]

        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(self.apt)
        self.assertEqual('vio_id2', vio_ret_uuid)
        self.assertEqual('vg4', vg_ret_uuid)

        mock_vios_get.reset_mock()
        mock_vg_get.reset_mock()

        # 2: At this point, the statics are set.  If we validate again, and the
        # VG.get returns the right one, we should bail out early.
        mock_vg_get.side_effect = None
        mock_vg_get.return_value = vg4

        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(self.apt)
        self.assertEqual('vio_id2', vio_ret_uuid)
        self.assertEqual('vg4', vg_ret_uuid)

        # Statics unchanged
        self.assertEqual('vg4', vopt._cur_vg_uuid)
        self.assertEqual('vio_id2', vopt._cur_vios_uuid)

        # We didn't have to query the VIOS
        mock_vios_get.assert_not_called()
        # We only did VG.get once
        self.assertEqual(1, mock_vg_get.call_count)

        mock_vg_get.reset_mock()

        # 3: Same again, but this time the repos is somewhere else.  We should
        # find it.
        vg4.vmedia_repos = []
        vg2.vmedia_repos = [1]
        # The first VG.get is looking for the already-set repos.  The second
        # will be the feed from the first VIOS.  There should be no third call,
        # since we should find the repos on VIOS 2.
        mock_vg_get.side_effect = [vg4, [vg1, vg2]]

        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(self.apt)
        self.assertEqual('vio_id1', vio_ret_uuid)
        self.assertEqual('vg2', vg_ret_uuid)

        # And the static values
        self.assertEqual('vg2', vopt._cur_vg_uuid)
        self.assertEqual('vio_id1', vopt._cur_vios_uuid)

        mock_vg_get.reset_mock()
        mock_vios_get.reset_mock()

        # 4: No repository anywhere - need to create one.  The default VG name
        # (rootvg) exists in multiple places.  Ensure we create in the first
        # one, for efficiency.
        vg2.vmedia_repos = []
        mock_vg_get.side_effect = [vg1, [vg1, vg2], [vg3, vg4]]
        vg1.update.return_value = vg1

        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(self.apt)
        self.assertEqual('vio_id1', vio_ret_uuid)
        self.assertEqual('vg1', vg_ret_uuid)

        self.assertEqual('vg1', vopt._cur_vg_uuid)
        self.assertEqual('vio_id1', vopt._cur_vios_uuid)
        self.assertEqual([mock_vmr_bld.return_value], vg1.vmedia_repos)

        mock_vg_get.reset_mock()
        mock_vios_get.reset_mock()
        vg1 = mock.MagicMock()

        # 5: No repos, need to create one.  But not on the mgmt partition.
        vwrap1.configure_mock(name='vio1', rmc_state='active', uuid='vio_id1',
                              is_mgmt_partition=True)
        vg3.vmedia_repos = []
        mock_vg_get.side_effect = [vg1, [vg1, vg2], [vg3, vg4]]
        vg3.update.return_value = vg3

        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(self.apt)
        self.assertEqual('vio_id2', vio_ret_uuid)
        self.assertEqual('vg3', vg_ret_uuid)

        self.assertEqual('vg3', vopt._cur_vg_uuid)
        self.assertEqual('vio_id2', vopt._cur_vios_uuid)
        self.assertEqual([mock_vmr_bld.return_value], vg3.vmedia_repos)

        mock_vg_get.reset_mock()
        mock_vios_get.reset_mock()
        vg3 = mock.MagicMock()

        # 6: No repos, and a configured VG name that doesn't exist
        vwrap1.configure_mock(name='vio1', rmc_state='active', uuid='vio_id1',
                              is_mgmt_partition=False)
        vg4.vmedia_repos = []
        mock_vg_get.side_effect = [vg1, [vg1, vg2], [vg3, vg4]]

        self.assertRaises(pvm_ex.NoMediaRepoVolumeGroupFound,
                          vopt.validate_vopt_repo_exists, self.apt,
                          vopt_media_volume_group='mythicalvg')

        # 7: No repos - need to create.  Make sure conf setting is honored.
        vg1.vmedia_repos = []

        mock_vg_get.side_effect = [vg1, [vg1, vg2], [vg3, vg4]]
        vg4.update.return_value = vg4

        vio_ret_uuid, vg_ret_uuid = vopt.validate_vopt_repo_exists(
            self.apt, vopt_media_volume_group='other2vg')
        self.assertEqual('vio_id2', vio_ret_uuid)
        self.assertEqual('vg4', vg_ret_uuid)

        self.assertEqual('vg4', vopt._cur_vg_uuid)
        self.assertEqual('vio_id2', vopt._cur_vios_uuid)
        self.assertEqual([mock_vmr_bld.return_value], vg4.vmedia_repos)
        vg1.update.assert_not_called()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

import unittest

from pypowervm.tasks.hdisk import _fc as fc
import pypowervm.tests.tasks.util as tju

VIOS_FEED = 'fake_vios_feed.txt'


class TestFC(unittest.TestCase):

    def setUp(self):
        super(TestFC, self).setUp()

    def test_itl(self):
        """Tests the ITL class."""
        itl = fc.ITL('AABBCCDDEEFF0011', '00:11:22:33:44:55:66:EE', 238)
        self.assertEqual('aabbccddeeff0011', itl.initiator)
        self.assertEqual('00112233445566ee', itl.target)
        self.assertEqual('ee000000000000', itl.lun)
        # Test Lun-ID length for max 8 bytes
        itl = fc.ITL('AABBC11', '00:11:22:33:44:55:66:EE', 1074872357)
        self.assertEqual('4011402500000000', itl.lun)
        self.assertEqual(16, len(itl.lun))

    def test_build_itls(self):
        """Tests that the ITL combinations can be built out."""
        i_wwpns = ['0011223344556677', '0011223344556678']
        t_wwpns = ['1111223344556677', '1111223344556678', '1111223344556679']
        all_itls = fc.build_itls(i_wwpns, t_wwpns, 238)

        self.assertEqual(6, len(all_itls))

        combos = [fc.ITL(i_wwpns[0], t_wwpns[0], 238),
                  fc.ITL(i_wwpns[0], t_wwpns[1], 238),
                  fc.ITL(i_wwpns[0], t_wwpns[2], 238),
                  fc.ITL(i_wwpns[1], t_wwpns[0], 238),
                  fc.ITL(i_wwpns[1], t_wwpns[1], 238),
                  fc.ITL(i_wwpns[1], t_wwpns[2], 238)]
        self.assertListEqual(combos, all_itls)

    def test_lua_recovery_xml(self):
        """Validates that the LUA recovery XML is build properly."""
        i_wwpns = ['0011223344556677', '0011223344556678']
        t_wwpns = ['1111223344556677', '1111223344556678', '1111223344556679']
        all_itls = fc.build_itls(i_wwpns, t_wwpns, 238)

        lua_xml = ('<XML_LIST><general><cmd_version>3</cmd_version><version>'
                   '2.0</version></general><reliableITL>false</reliableITL>'
                   '<deviceList><device><vendor>OTHER</vendor><deviceTag>'
                   '1</deviceTag><itlList><number>6</number><itl>'
                   '<Iwwpn>0011223344556677</Iwwpn><Twwpn>1111223344556677'
                   '</Twwpn><lua>ee000000000000</lua></itl><itl><Iwwpn>'
                   '0011223344556677</Iwwpn><Twwpn>1111223344556678</Twwpn>'
                   '<lua>ee000000000000</lua></itl><itl><Iwwpn>'
                   '0011223344556677</Iwwpn><Twwpn>1111223344556679</Twwpn>'
                   '<lua>ee000000000000</lua></itl><itl><Iwwpn>'
                   '0011223344556678</Iwwpn><Twwpn>1111223344556677</Twwpn>'
                   '<lua>ee000000000000</lua></itl><itl><Iwwpn>'
                   '0011223344556678</Iwwpn><Twwpn>1111223344556678</Twwpn>'
                   '<lua>ee000000000000</lua></itl><itl><Iwwpn>'
                   '0011223344556678</Iwwpn><Twwpn>1111223344556679</Twwpn>'
                   '<lua>ee000000000000</lua></itl></itlList></device>'
                   '</deviceList></XML_LIST>')

        self.assertEqual(lua_xml, fc._lua_recovery_xml(all_itls, None))

    def test_process_lua_result_no_resp(self):
        result = {}
        status, dev_name, udid = fc._process_lua_result(result)
        self.assertIsNone(status)
        self.assertIsNone(dev_name)
        self.assertIsNone(udid)

    def test_process_lua_result_terse_resp(self):
        """Tests where valid XML is returned, but no device."""
        xml = ('<luaResult><version>2.0</version><deviceList></deviceList>'
               '</luaResult>')
        result = {'StdOut': xml}
        status, dev_name, udid = fc._process_lua_result(result)
        self.assertIsNone(status)
        self.assertIsNone(dev_name)
        self.assertIsNone(udid)

    def test_process_lua_result(self):
        xml = ('<luaResult><version>2.0</version><deviceList><device>'
               '<deviceTag>21</deviceTag><status>8</status><msg><msgLen>9'
               '</msgLen><msgText>test text</msgText></msg><pvName>hdisk10'
               '</pvName><uniqueID>fake_uid</uniqueID><udid>fake_udid'
               '</udid></device></deviceList></luaResult>')
        result = {'StdOut': xml}
        status, dev_name, udid = fc._process_lua_result(result)
        self.assertEqual('8', status)
        self.assertEqual('hdisk10', dev_name)
        self.assertEqual('fake_udid', udid)

        # Repeat with the input as the resultXML
        result = {'OutputXML': xml}
        status, dev_name, udid = fc._process_lua_result(result)
        self.assertEqual('8', status)
        self.assertEqual('hdisk10', dev_name)
        self.assertEqual('fake_udid', udid)

    @mock.patch('pypowervm.tasks.hdisk._fc.LOG')
    def test_validate_lua_status(self, mock_log):
        """This tests the branches of validate_lua_status."""
        fc._log_lua_status(fc.LUAStatus.DEVICE_AVAILABLE, 'dev_name',
                           'message')
        self.assertEqual(1, mock_log.info.call_count)

        fc._log_lua_status(fc.LUAStatus.FOUND_ITL_ERR, 'dev_name', 'message')
        self.assertEqual(1, mock_log.warning.call_count)

        fc._log_lua_status(fc.LUAStatus.DEVICE_IN_USE, 'dev_name', 'message')
        self.assertEqual(2, mock_log.warning.call_count)

        fc._log_lua_status(fc.LUAStatus.FOUND_DEVICE_UNKNOWN_UDID, 'dev_name',
                           'message')
        self.assertEqual(3, mock_log.warning.call_count)

        fc._log_lua_status(fc.LUAStatus.INCORRECT_ITL, 'dev_name', 'message')
        self.assertEqual(4, mock_log.warning.call_count)

    @mock.patch('pypowervm.tasks.hdisk._fc._process_lua_result')
    @mock.patch('pypowervm.wrappers.job.Job', new=mock.Mock())
    @mock.patch('pypowervm.adapter.Adapter')
    def test_lua_recovery(self, mock_adapter, mock_lua_result):
        itls = [fc.ITL('AABBCCDDEEFF0011', '00:11:22:33:44:55:66:EE', 238)]

        mock_lua_result.return_value = ('OK', 'hdisk1', 'udid')

        status, devname, udid = fc.lua_recovery(mock_adapter, 'vios_uuid',
                                                itls)

        # Validate value unpack
        self.assertEqual('OK', status)
        self.assertEqual('hdisk1', devname)
        self.assertEqual('udid', udid)

        # Validate method invocations
        self.assertEqual(1, mock_adapter.read.call_count)
        self.assertEqual(1, mock_lua_result.call_count)

    @mock.patch('pypowervm.tasks.hdisk._fc._lua_recovery_xml')
    @mock.patch('pypowervm.tasks.hdisk._fc._process_lua_result')
    @mock.patch('pypowervm.wrappers.job.Job', new=mock.Mock())
    @mock.patch('pypowervm.adapter.Adapter')
    def test_lua_recovery_dupe_itls(self, mock_adapter, mock_lua_result,
                                    mock_lua_xml):
        itls = [fc.ITL('AABBCCDDEEFF0011', '00:11:22:33:44:55:66:EE', 238),
                fc.ITL('AABBCCDDEEFF0011', '00:11:22:33:44:55:66:EE', 238)]

        mock_lua_result.return_value = ('OK', 'hdisk1', 'udid')

        status, devname, udid = fc.lua_recovery(mock_adapter, 'vios_uuid',
                                                itls)

        # Validate value unpack
        self.assertEqual('OK', status)
        self.assertEqual('hdisk1', devname)
        self.assertEqual('udid', udid)

        # Validate method invocations
        self.assertEqual(1, mock_adapter.read.call_count)
        self.assertEqual(1, mock_lua_result.call_count)
        mock_lua_xml.assert_called_with({itls[0]}, mock_adapter,
                                        vendor='OTHER')

    @mock.patch('pypowervm.tasks.hdisk._fc.lua_recovery')
    @mock.patch('pypowervm.utils.transaction.FeedTask')
    @mock.patch('pypowervm.tasks.storage.add_lpar_storage_scrub_tasks')
    @mock.patch('pypowervm.tasks.storage.find_stale_lpars')
    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.get',
                new=mock.Mock())
    def test_discover_hdisk(self, mock_fsl, mock_alsst, mock_ftsk, mock_luar):
        def set_luar_side_effect(_stat, _dev):
            """Set up the lua_recovery mock's side effect.

            The second return will always be the same - used to verify that we
            really called twice when appropriate.
            The first return will be (_stat, _dev, "udid"), per the params.
            """
            mock_luar.reset_mock()
            mock_luar.side_effect = [(_stat, _dev, 'udid'),
                                     ('ok_s', 'ok_h', 'ok_u')]
        stale_lpar_ids = [12, 34]
        # All of these should cause a scrub-and-retry
        retry_rets = [(None, None), (fc.LUAStatus.DEVICE_AVAILABLE, None),
                      (fc.LUAStatus.FOUND_DEVICE_UNKNOWN_UDID, 'hdisk456')]
        # These should *not* cause a scrub-and-retry
        no_retry_rets = [(fc.LUAStatus.DEVICE_AVAILABLE, 'hdisk456'),
                         (fc.LUAStatus.FOUND_ITL_ERR, 'hdisk456'),
                         (fc.LUAStatus.DEVICE_IN_USE, 'hdisk456')]
        mock_fsl.return_value = stale_lpar_ids
        for st, dev in retry_rets:
            set_luar_side_effect(st, dev)
            self.assertEqual(
                ('ok_s', 'ok_h', 'ok_u'), fc.discover_hdisk(
                    'adp', 'vuuid', ['itls']))
            self.assertEqual(1, mock_fsl.call_count)
            mock_ftsk.assert_called_with('scrub_vios_vuuid', mock.ANY)
            self.assertEqual(1, mock_alsst.call_count)
            mock_luar.assert_has_calls(
                [mock.call('adp', 'vuuid', ['itls'],
                           vendor=fc.LUAType.OTHER)] * 2)
            mock_fsl.reset_mock()
            mock_alsst.reset_mock()
            mock_ftsk.reset_mock()

        for st, dev in no_retry_rets:
            set_luar_side_effect(st, dev)
            self.assertEqual(
                (st, dev, 'udid'), fc.discover_hdisk(
                    'adp', 'vuuid', ['itls']))
            self.assertEqual(0, mock_fsl.call_count)
            self.assertEqual(0, mock_ftsk.call_count)
            self.assertEqual(0, mock_alsst.call_count)
            self.assertEqual(1, mock_luar.call_count)
            mock_luar.assert_called_with('adp', 'vuuid', ['itls'],
                                         vendor=fc.LUAType.OTHER)

        # If no stale LPARs found, scrub-and-retry should not be triggered with
        # either set.
        mock_fsl.return_value = []
        for st, dev in retry_rets + no_retry_rets:
            set_luar_side_effect(st, dev)
            self.assertEqual(
                (st, dev, 'udid'), fc.discover_hdisk(
                    'adp', 'vuuid', ['itls']))
            # find_stale_lpars will be called for retry_rets, but not for
            # no_retry_rets
            self.assertLessEqual(mock_fsl.call_count, 1)
            self.assertEqual(0, mock_ftsk.call_count)
            self.assertEqual(0, mock_alsst.call_count)
            self.assertEqual(1, mock_luar.call_count)
            mock_luar.assert_called_with('adp', 'vuuid', ['itls'],
                                         vendor=fc.LUAType.OTHER)
            mock_fsl.reset_mock()

    @mock.patch('pypowervm.wrappers.job.Job.job_status', new=mock.Mock())
    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    @mock.patch('pypowervm.adapter.Adapter')
    def test_remove_hdisk_classic(self, mock_adapter, mock_run_job):
        mock_adapter.read.return_value = (tju.load_file(VIOS_FEED)
                                          .feed.entries[0])

        fc._remove_hdisk_classic(mock_adapter, 'host_name', 'dev_name',
                                 'vios_uuid')
        # Validate method invocations
        self.assertEqual(2, mock_adapter.read.call_count)
        self.assertEqual(1, mock_run_job.call_count)

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    @mock.patch('pypowervm.adapter.Adapter')
    def test_remove_hdisk_job(self, mock_adapter, mock_run_job):
        mock_adapter.read.return_value = (tju.load_file(VIOS_FEED)
                                          .feed.entries[0])

        def verify_run_job(vios_uuid, job_parms=None):
            self.assertEqual('vios_uuid', vios_uuid)
            self.assertEqual(1, len(job_parms))
            job_parm = (b'<web:JobParameter xmlns:web="http://www.ibm.com/'
                        b'xmlns/systems/power/firmware/web/mc/2012_10/" '
                        b'schemaVersion="V1_0"><web:ParameterName>devName'
                        b'</web:ParameterName><web:ParameterValue>dev_name'
                        b'</web:ParameterValue></web:JobParameter>')
            self.assertEqual(job_parm, job_parms[0].toxmlstring())

        mock_run_job.side_effect = verify_run_job

        fc._remove_hdisk_job(mock_adapter, 'dev_name', 'vios_uuid')
        # Validate method invocations
        self.assertEqual(1, mock_adapter.read.call_count)
        self.assertEqual(1, mock_run_job.call_count)

    def test_normalize_lun(self):
        lun = fc.normalize_lun(12)
        self.assertEqual('c000000000000', lun)
        # Test when lun exceeds len 8
        lun = fc.normalize_lun(1074872357)
        self.assertEqual('4011402500000000', lun)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes for all wrapper classes in the pypowervm.wrappers package."""

import abc
from oslo_log import log as logging
import re
import six
import urllib


from pypowervm import adapter as adpt
import pypowervm.const as pc
import pypowervm.entities as ent
from pypowervm.i18n import _
from pypowervm import util
import pypowervm.utils.uuid as pvm_uuid

LOG = logging.getLogger(__name__)


def _indirect_child_elem(wrap, indirect):
    if indirect is None:
        return wrap.element
    else:
        return ent.Element(indirect, wrap.adapter, children=[wrap.element])


@six.add_metaclass(abc.ABCMeta)
class Wrapper(object):
    """Base wrapper object that subclasses should extend.

    Provides base support for operations.  Will define a few methods that need
    to be overridden by subclasses.
    """

    # See pvm_type decorator
    schema_type = None
    default_attrib = None
    schema_ns = None
    has_metadata = False

    # Registers PowerVM object wrappers by their schema type.
    # {schema_type_string:
    #     {'entry': entry_wrapper_class,
    #      'element': element_wrapper_class},
    #  ...}
    # E.g. {'SharedEthernetAdapter':
    #          {'entry': pypowervm.wrappers.network.SEA},
    #       'LogicalUnit':
    #          {'entry': pypowervm.wrappers.storage.LUEnt,
    #           'element': pypowervm.wrappers.storage.LU},
    #       ...}
    # Some schema types double as a ROOT/CHILD object and a DETAIL within a
    # ROOT/CHILD).  Such schema types will have both 'entry' and 'element' keys
    # in the sub-dict.  Otherwise, only one key will exist accordingly.
    _pvm_object_registry = {}

    # Maps a property name to its extended attribute group.  Should be sparse -
    # if a property is not associated with a xag, it can (should) be absent
    # from this dict.
    _xag_registry = {}

    # Allows us to ensure that all Wrapper classes are properly registered via
    # @[base_]pvm_type.
    _registered = False

    @classmethod
    def base_pvm_type(cls, cls_):
        """Decorator/method to register a PowerVM base class.

        Use this instead of @pvm_type on Wrapper subclasses which are not to be
        instantiated, but are themselves bases for real Wrappers.  For example,
        use @base_pvm_type for BasePartition; and @pvm_type for
        LogicalPartition and VirtualIOServer.

        Use as a decorator with no arguments:

        @Wrapper.base_pvm_type
        class SomeBaseClass(Wrapper):
            ...

        Or use as a method to register a base class explicitly after it has
        been defined:

        Wrapper.base_pvm_type(SomeBaseClass)

        :param cls_: The Wrapper subclass to be decorated/registered.
        :return: cls_
        """
        # @xag_property registers with Wrapper._xag_registry because
        # cls_ hasn't been created yet.  Transfer the created registry to the
        # cls_, merging with any already registered by its bases, and clear
        # Wrapper's registry so it doesn't pollute the next cls_.
        cls_._xag_registry = dict(cls_._xag_registry, **Wrapper._xag_registry)
        cls_._registered = True
        Wrapper._xag_registry = {}
        return cls_

    @classmethod
    def _register_schema_type(cls, schema_type, class_):
        """Register this class according to its schema_type.

        The registry is used to identify the appropriate wrapper class to apply
        to an XML object we receive from the REST server.

        :param schema_type: String schema type of the REST element to register.
        :param class_: Concrete {Entry|Element}Wrapper subclass to associate
                       with the schema_type.
        """
        ent_or_el = 'entry' if issubclass(class_, EntryWrapper) else 'element'
        if schema_type not in Wrapper._pvm_object_registry:
            Wrapper._pvm_object_registry[schema_type] = {}
        Wrapper._pvm_object_registry[schema_type][ent_or_el] = class_

    @classmethod
    def pvm_type(cls, schema_type, has_metadata=None, ns=pc.UOM_NS,
                 attrib=pc.DEFAULT_SCHEMA_ATTR, child_order=None):
        """Decorator for {Entry|Element}Wrappers of PowerVM objects.

        Sets foundational fields used for construction of new wrapper instances
        and pieces thereof.

        Registers the decorated class, keyed by its schema_type,  This enables
        the wrap method to return the correct subclass even if invoked directly
        from ElementWrapper or EntryWrapper.

        :param schema_type: PowerVM REST API Schema type of the subclass (str).
        :param has_metadata: Indicates whether, when creating and wrapping a
                             fresh adapter.Element, it should have a
                             <Metadata/> child element.
        :param ns: PowerVM REST API Schema namespace of the subclass.
        :param attrib: Default attributes for fresh Element when factory
                       constructor is used.
        :param child_order: Ordered list of the element names of the first-
                            level children of this element/entry.  Used for
                            order-agnostic construction/setting of values.
        """
        def inner(class_):
            # Base stuff first (e.g. register extended attribute groups).
            cls.base_pvm_type(class_)
            class_.schema_type = schema_type
            if has_metadata is not None:
                class_.has_metadata = has_metadata
            if ns is not None:
                class_.schema_ns = ns
            if attrib is not None:
                class_.default_attrib = attrib
            if child_order:
                co = list(child_order)
                if class_.has_metadata and co[0] != 'Metadata':
                    co.insert(0, 'Metadata')
                class_._child_order = tuple(co)
            cls._register_schema_type(schema_type, class_)
            return class_
        return inner

    @classmethod
    def xag_property(cls, xag):
        """Decorator to tag a @property with an extended attribute group.

        Use this decorator in place of (not in addition to) @property.  Within
        class Foo:

            @xag_property('bar')
            def some_prop(self):
                ...

        confers the same property-ness on 'some_prop' as would

            @property
            def some_prop(self):
                ...

        but it also associates some_prop with extended attribute group name
        'bar' such that Foo.get_xag_for_prop('some_prop') returns the value
        'bar'.

        :param xag: String name of the extended attribute group with which the
                    decorated property is associated.  May either be one of the
                    pypowervm.const.XAG enum values; or a member of one of the
                    pypowervm.entities.*XAGs collections (for example, see
                    pypowervm.wrappers.virtual_io_server.phys_vols).
        """
        def wrap(func):
            cls._xag_registry[func.__name__] = str(xag)
            return property(func)
        return wrap

    @classmethod
    def get_xag_for_prop(cls, propname):
        """The extended attribute group name for a property of this Wrapper.

        :param propname: Short (unqualified) name of a property of this
                         Wrapper, as a string.
        :return: String indicating the name of the extended attribute group for
                 the given property.  Should be a pypowervm.const.XAG enum
                 value.  None (not 'None') if there is no xag associated with
                 the specified property.
        """
        return cls._xag_registry.get(propname, None)

    @property
    def child_order(self):
        return getattr(self, '_child_order', ())

    @property
    def adapter(self):
        return self.element.adapter

    @property
    def traits(self):
        return self.adapter.traits

    @property
    def uuid(self):
        """Returns the uuid of the entry or element."""
        # The following should only apply to EntryWrappers
        if getattr(self, 'entry', None) is not None:
            return self.entry.uuid

        # Anything with Metadata may have a UUID.  Could do has_metadata check,
        # but that doesn't really add anything.  This will return None if not
        # found.
        return self._get_val_str(pc.UUID_XPATH)

    @uuid.setter
    def uuid(self, value):
        """Sets the UUID (if supported).

        :param value: A valid PowerVM UUID value in either uuid format
                      or string format
        """
        if isinstance(self, WrapperSetUUIDMixin):
            self.set_uuid(value)
        else:
            raise AttributeError(_('Cannot set uuid.'))

    def inject(self, subelement, replace=True):
        """Injects subelement as a child element, possibly replacing it.

        This is pypowervm.adapter.Element.inject, with ordering_list always set
        to self.child_order.
        """
        self.element.inject(subelement, self.child_order, replace=replace)

    def _find(self, property_name, use_find_all=False):
        """Will find a given element within the object.

        :param property_name: The property to search within the tree for.
        :param use_find_all: If set to true, will use the find_all method for
                             queries.
        """
        element = self.element
        if element is None:
            return None
        if use_find_all:
            found_value = element.findall(property_name)
        else:
            found_value = element.find(property_name)

        return found_value  # May be None

    def _find_or_seed(self, prop_name, attrib=pc.DEFAULT_SCHEMA_ATTR):
        """Will find the existing element, or create if needed.

        If the element is not found, it will be added to the child list
        of this element.
        :param prop_name: The property name to replace with the new value.
        :param attrib: The attributes to use for the property.  Defaults to
                       the DEFAULT_SCHEM_ATTR.
        :returns: The existing element, or a newly created one if not found.
        """
        root_elem = self.element

        # Find existing
        existing = root_elem.find(prop_name)
        if existing:
            return existing
        else:
            new_elem = ent.Element(prop_name, self.adapter, attrib=attrib,
                                   children=[])
            self.inject(new_elem)
            return new_elem

    def _get_elem_list(self, tag):
        """An entities.ElementList for a given tag from within this wrapper.

        :param tag: The string XML tag of the values to find.
        :return: An entities.ElementList for the specified tag.
        """
        return ent.ElementList(
            self.element, tag, ordering_list=self.child_order)

    def _set_elem_list(self, tag, val_iter):
        """Set (or replace) the contents of an entities.ElementList.

        :param tag: The string XML tag of the ElementList to assign.
        :param val_iter: Iterable of raw (string) values to set.
        """
        ellist = self._get_elem_list(tag)
        ellist.clear()
        ellist.extend(val_iter)

    def replace_list(self, prop_name, prop_children,
                     attrib=pc.DEFAULT_SCHEMA_ATTR, indirect=None):
        """Replaces a property on this Entry that contains a children list.

        The prop_children represent the new elements for the property.

        If the property does not already exist, this will simply append the
        new children.

        :param prop_name: The property name to replace with the new value.
        :param prop_children: A list of ElementWrapper objects that represent
                              the new children for the property list.
        :param attrib: The attributes to use if the property.  Defaults to
                       the DEFAULT_SCHEM_ATTR.
        :param indirect: Name of a schema element which should wrap each of the
                         prop_children.  For example, VNIC backing devices look
                         like:
                            <AssociatedBackingDevices>
                                <Metadata>...</Metadata>
                                <VirtualNICBackingDeviceChoice>
                                    <VirtualNICSRIOVBackingDevice>
                                        ...
                                    </VirtualNICSRIOVBackingDevice>
                                </VirtualNICBackingDeviceChoice>
                                <VirtualNICBackingDeviceChoice>
                                    <VirtualNICSRIOVBackingDevice>
                                        ...
                                    </VirtualNICSRIOVBackingDevice>
                                </VirtualNICBackingDeviceChoice>
                                ...
                            </AssociatedBackingDevices>
                         In this case, invoke this method as:
                         replace_list(
                             'AssociatedBackingDevices',
                             [<list of VirtualNICSRIOVBackingDevice wrappers>],
                             indirect='VirtualNICBackingDeviceChoice')
        """
        new_elem = ent.Element(prop_name, self.adapter, attrib=attrib,
                               children=[_indirect_child_elem(child, indirect)
                                         for child in prop_children])

        self.inject(new_elem)

    @abc.abstractproperty
    def _type_and_uuid(self):
        """Return the type and uuid of this entry together in one string.

        This is useful for error messages, logging, etc.
        """
        pass

    def set_parm_value(self, property_name, value, create=True, attrib=None):
        """Set a child element value, possibly creating the child.

        :param property_name: The schema name of the property to set.
        :param value: The (string) value to assign to the property's 'text'.
        :param create: If True, and the property is not found, it will be
        created.  Otherwise this method will throw an exception.
        :param attrib: The element attributes to use if the element is created.
        """
        element_value = self._find(property_name)
        if element_value is None:
            self.log_missing_value(property_name)
            if create:
                element_value = ent.Element(
                    property_name, self.adapter, ns=self.schema_ns,
                    attrib=attrib, text=str(value))
                self.inject(element_value)

        element_value.text = str(value)

    def set_float_gb_value(self, property_name, value, create=True):
        """Special case of set_parm_value for floats of Gigabyte.Type.

        - Gigabyte.Type can't handle more than 6dp.
        - Floating point representation issues can mean that e.g. 0.1 + 0.2
        produces 0.30000000000000004.
        - str() rounds to 12dp.

        So this method converts a float (or float string) to a string with
        exactly 6dp before storing it in the property.

        :param property_name: The schema name of the property to set (see
                              set_parm_value).
        :param value: The floating point number or floating point string to be
                      set.
        :param create: If True, and the property is not found, it will be
                       created.  Otherwise this method will throw an exception.
                       (See set_parm_value.)
        """
        self.set_parm_value(property_name,
                            util.sanitize_float_for_api(value, precision=6),
                            create=create)

    def __get_val(self, property_name, default=None, converter=None):
        """Retrieve the value of an element within this wrapper's ElementTree.

        This is the baseline for all the _get_val_{type} methods.
        :param property_name: The name (XPath) of the property to find.
        :param default: The default value to return if the property is not
                        found OR if type conversion fails.
        :param converter: Optional callable accepting a single string parameter
                          and returning a value of some other type.  The
                          converter callable should raise ValueError if
                          conversion fails.
        :return: The (possibly converted) value corresponding to the identified
                 property.
        """
        element_value = self._find(property_name)
        if element_value is None:
            self.log_missing_value(property_name)
            return default

        text = element_value.text
        if text is None:
            return default

        if type(text) is str:
            text = text.strip()

        if callable(converter):
            try:
                return converter(text)
            except ValueError:
                message = (_(
                    "Cannot convert %(property_name)s='%(value)s' in object "
                    "%(pvmobject)s") % {"property_name": property_name,
                                        "value": text,
                                        "pvmobject": self._type_and_uuid})

                LOG.error(message)
                return default
        return text

    def _get_vals(self, property_name):
        """Gets a list of values from PowerVM.

        :param property_name: property to return
        :returns: List of strings containing property values.  No type
                  conversion is done.  If no elements are found, the empty list
                  is returned (as opposed to None).
        """
        values = []
        elements = self._find(property_name, use_find_all=True)
        if elements is not None:
            for element in elements:
                values.append(element.text)
        return values

    def _get_val_bool(self, property_name, default=False):
        """Gets the boolean value of a PowerVM property.

        :param property_name: property to return
        :param default: The value to return if the property is not found
            in the data
        :return: If the property exists in the data and has the value
            'true' or 'false' (case-insensitive), then the corresponding
            boolean value will be returned.
            If the property does not exist, then the default value will be
            returned if specified, otherwise False will be returned.
        """
        def str2bool(bool_str):
            return str(bool_str).lower() == 'true'
        return self.__get_val(property_name, default=default,
                              converter=str2bool)

    def _get_val_int(self, property_name, default=None):
        """Gets the integer value of a PowerVM property.

        :param property_name: property to find
        :param default: Value to return if property is not found.  Defaults to
                        None (which is not an int - plan accordingly).
        :return: Integer (int) value of the property if it is found and it is a
                 valid integer.
        :raise ValueError: If the value cannot be converted.
        """
        return self.__get_val(property_name, default=default, converter=int)

    def _get_val_float(self, property_name, default=None):
        """Gets the float value of a PowerVM property.

        :param property_name: property to find
        :param default: Value to return if property is not found.  Defaults to
                        None (which is not a float - plan accordingly).
        :return: float value of the property if it is found and it is a
                 valid float.
        :raise ValueError: If the value cannot be converted.
        """
        return self.__get_val(property_name, default=default, converter=float)

    def _get_val_str(self, property_name, default=None):
        """Gets the string value of a PowerVM property.

        :param property_name: property to find
        :param default: Value to return if property is not found.  Defaults to
                        None (which is not a str - plan accordingly).
        :return: str value of the property if it is found.  May be the empty
                 string.
        """
        return self.__get_val(property_name, default=default, converter=None)

    def _get_val_percent(self, property_name, default=None):
        """Gets the value in float-percentage format of a PowerVM property.

        :param property_name: property to find
        :param default: Value to return if property is not found. Defaults to
                        None (which is not a float - plan accordingly).
        :return: If the property is say "2.45%", a value of .0245 will be
                 returned. % in the property is optional.
        """
        def str2percent(percent_str):
            if percent_str:
                percent_str = re.findall(r"\d*\.?\d+", percent_str)[0]
                return (float(percent_str))/100
            else:
                return None
        return self.__get_val(property_name, default=default,
                              converter=str2percent)

    def log_missing_value(self, param):
        LOG.trace('The expected parameter of %(param)s was not found in '
                  '%(identifier)s' % {"param": param,
                                      "identifier": self._type_and_uuid})

    def get_href(self, propname, one_result=False):
        """Returns the hrefs from AtomLink elements.

        :param propname: The name of the schema element containing the 'href'
                         attribute.
        :param one_result: If True, we are expecting exactly one result, and
                           will return that one (string) result, or None.  If
                           False (the default), we will return a tuple of
                           strings which may be empty.
        """
        ret_links = []
        elements = self._find(propname, use_find_all=True)

        # Loop through what was found, if anything
        if elements:
            for atomlink in elements:
                # If the element doesn't have an href, ignore it.
                try:
                    ret_links.append(atomlink.attrib['href'])
                except KeyError:
                    pass

        if one_result:
            if len(ret_links) == 1:
                return ret_links[0]
            else:
                return None
        # Otherwise return a (possibly empty) tuple of the results
        return tuple(ret_links)

    def set_href(self, propname, href):
        """Finds or creates the (single) named property and sets its href.

        If the indicated element does not exist, it (and any necessary interim
        parent elements) will be created.  If any intervening path is non-
        unique, any new element paths will be created under the first one.

        :param propname: XPath to the property.
        :param href: The URI value to assign to the href attribute.
                     rel=related is automatically assigned.
        """
        links = self._find(propname, use_find_all=True)
        if len(links) > 1:
            msg = _('Refusing set href over multiple links.\nPath: %{path}s\n'
                    'Number of links found: %{nlinks}d')
            raise ValueError(msg % {'path': propname, 'nlinks': len(links)})
        if len(links) == 1:
            link = links[0]
        else:
            # Not found - create the property
            pathtoks = propname.split(util.XPATH_DELIM)
            append_point = self
            while len(pathtoks) > 1:
                next_prop = pathtoks.pop(0)
                new_el = append_point._find(next_prop)
                if new_el is None:
                    new_el = ent.Element(next_prop, self.adapter)
                    append_point.element.inject(
                        new_el, ordering_list=self.child_order)
                append_point = ElementWrapper.wrap(new_el)
            link = ent.Element(pathtoks[-1], self.adapter)
            append_point.element.inject(link, ordering_list=self.child_order)
        # At this point we have found or created the propname element.  Its
        # handle is in the link var.
        link.attrib['href'] = href
        link.attrib['rel'] = 'related'

    def toxmlstring(self, pretty=False):
        """Produce an XML dump of this Wrapper's Element.

        :param pretty: If True, format the XML in a visually-pleasing manner.
        :return: An XML string representing this Element.
        """
        return self.element.toxmlstring(pretty=pretty)

    @classmethod
    def _bld_element(cls, adapter, tag=None, has_metadata=has_metadata,
                     ns=schema_ns, attrib=default_attrib):
        """Create a fresh entities.Element, usually for immediate wrapping.

        :param adapter: The entities.Adapter to be consulted for traits, etc.
        :param tag: Property name of the new Element.
        :param has_metadata: If True, a <Metadata/> child will be created.
        :param ns: Namespace to use.
        :param attrib: XML attributes to use in the outer Element.
        :return: A fresh adapter.Element.
        """
        # TODO(efried): Get rid of this method - fold it into Element._bld()
        tag = cls.schema_type if tag is None else tag
        # Make sure the call was either through a legal wrapper or explicitly
        # specified a tag name
        if tag is None:
            raise TypeError(_("Refusing to construct and wrap an Element "
                              "without a tag."))
        has_metadata = (cls.has_metadata
                        if has_metadata is None
                        else has_metadata)
        ns = cls.schema_ns if ns is None else ns
        attrib = cls.default_attrib if attrib is None else attrib
        children = []
        if has_metadata:
            children.append(
                ent.Element('Metadata', adapter, ns=ns, children=[
                    ent.Element('Atom', adapter, ns=ns)]))
        return ent.Element(tag, adapter, ns=ns, attrib=attrib,
                           children=children)

    @classmethod
    def _class_for_element(cls, element):
        """Discover and return an appropriate *Wrapper subclass for element.

        :param element: An adapter.Element to introspect
        :return: A Wrapper subclass (the class, not an instance).  If element
                 represents a known subclass, it is returned; else the invoking
                 class is returned.
        """
        # Extract tag.  Let this raise AttributeError - means the Element is
        # invalid.
        schema_type = element.tag
        # Is it a registered wrapper class?
        try:
            return Wrapper._pvm_object_registry[schema_type][
                'entry' if issubclass(cls, EntryWrapper) else 'element']
        except KeyError:
            return cls

    def _bld_link_list(self, container_type, links):
        """Creates an element with a list of <link> children.

        :param container_type: The element that will contain the elements.
        :param links: The set of strings which are link elements.
        """
        new_elems = []
        for item in links:
            new_elems.append(ent.Element('link', self.adapter, attrib={
                'href': item, 'rel': 'related'}))
        return ent.Element(container_type, self.adapter, children=new_elems)


class EntryWrapper(Wrapper):
    """Base Wrapper for the Entry object types."""
    # If it's an Entry, it must be a ROOT or CHILD
    has_metadata = True

    def __init__(self, entry, etag=None):
        self.entry = entry
        self._etag = etag

    @classmethod
    def getter(cls, adapter, entry_uuid=None, parent_class=None,
               parent_uuid=None, xag=None, parent=None):
        """Return EntryWrapperGetter or FeedGetter for this EntryWrapper type.

        Parameters are the same as described by EntryWrapperGetter.__init__
        If entry_uuid is None, a FeedGetter is returned.  Otherwise, an
        EntryWrapperGetter is returned.
        """
        if entry_uuid is None:
            return FeedGetter(
                adapter, cls, parent=parent, parent_class=parent_class,
                parent_uuid=parent_uuid, xag=xag)
        else:
            return EntryWrapperGetter(
                adapter, cls, entry_uuid, parent=parent,
                parent_class=parent_class, parent_uuid=parent_uuid, xag=xag)

    @classmethod
    def _bld(cls, adapter, tag=None, has_metadata=None, ns=None, attrib=None):
        """Create a fresh EntryWrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param tag: XML tag for the EntryWrapper's Entry's root Element.
        :param has_metadata: If True, a basic <Metadata/> child is created
                             under the root element.
        :param ns: XML namespace for the contents.
        :param attrib: XML attributes for the root element.
        """
        element = cls._bld_element(
            adapter, tag, has_metadata=has_metadata, ns=ns, attrib=attrib)
        return cls(ent.Entry({'title': element.tag}, element.element, adapter))

    @classmethod
    def wrap(cls, response_or_entry, etag=None):
        """Creates an entry (or list) from an adapter.Response or Entry.

        If response is specified and is a feed, a list of EntryWrapper will be
        returned.  The entries within the feed are not guaranteed to have etags
        (e.g. from non-uom elements).

        Otherwise, a single EntryWrapper will be returned.  This is NOT a list.

        This method should usually be invoked from an EntryWrapper subclass
        decorated by Wrapper.pvm_type, and an instance of that subclass will be
        returned.

        If invoked directly from EntryWrapper, we attempt to detect whether an
        appropriate subclass exists based on the Entry's Element's tag.  If so,
        that subclass is used; otherwise a generic EntryWrapper is used.

        :param response_or_entry: The Response from an adapter.Adapter.read
                                  request, or an existing adapter.Entry to
                                  wrap.
        :returns: A list of wrappers if response_or_entry is a Response with a
                  Feed.  A single wrapper if response_or_entry is an Entry or a
                  Response with an Entry.
        """
        # Process Response if specified.  This recursively calls this method
        # with the entry(s) within the Response.
        if isinstance(response_or_entry, adpt.Response):
            if response_or_entry.entry is not None:
                return cls.wrap(
                    response_or_entry.entry,
                    etag=response_or_entry.etag)
            elif response_or_entry.feed is not None:
                return [cls.wrap(entry, etag=entry.etag)
                        for entry in response_or_entry.feed.entries]
            else:
                raise KeyError(_("Response is missing 'entry' property."))

        # Else process Entry if specified
        if isinstance(response_or_entry, ent.Entry):
            # If schema_type is set, cls represents a legal subclass - use it.
            # Otherwise, try to discover an appropriate subclass based on the
            # element.  If that fails, it will default to the invoking class,
            # which will usually just be EntryWrapper.
            wcls = (cls._class_for_element(response_or_entry.element)
                    if cls.schema_type is None
                    else cls)
            return wcls(response_or_entry, etag)

        # response_or_entry is neither a Response nor an Entry
        fmt = _("Must supply a Response or Entry to wrap.  Got %s")
        raise TypeError(fmt % str(type(response_or_entry)))

    def refresh(self, use_etag=True):
        """Fetch the latest version of the entry from the REST API server.

        If the entry has not been updated on the server, self is returned
        unchanged.  Otherwise a new, fresh wrapper instance is returned.
        Generally, this should be used as:

        wrapper_instance = wrapper_instance.refresh()

        :param use_etag: (Optional) If False, the object's etag will not be
                         sent with the request, ensuring that the object is
                         retrieved afresh from the server.
        :return: EntryWrapper representing the latest data from the REST API
                 server.  If the input wrapper contains etag information and
                 the server responds 304 (Not Modified), the original wrapper
                 is returned.  Otherwise, a fresh EntryWrapper of the
                 appropriate type is returned.
        """
        etag = self.etag if use_etag else None
        resp = self.adapter.read_by_href(self.href, etag=etag)
        if resp.status == pc.HTTPStatus.NO_CHANGE:
            return self
        return self.wrap(resp)

    @classmethod
    def get(cls, adapter, uuid=None, parent_type=None, parent_uuid=None,
            parent=None, **read_kwargs):
        """GET and wrap an entry or feed of this type.

        Shortcut to EntryWrapper.wrap(adapter.read(...)).
        For example, retrieving a ROOT object:
            resp = adapter.read(VIOS.schema_type, root_id=v_uuid, xag=xags)
            vwrap = VIOS.wrap(resp)
        Becomes:
            vwrap = VIOS.get(adapter, uuid=v_uuid, xag=xags)

        Or retrieving a CHILD feed:
            resp = adapter.read(System.schema_type, root_id=sys_uuid,
                                child_type=VSwitch.schema_type)
            vswfeed = VSwitch.wrap(resp)
        Becomes:
            vswfeed = VSwitch.get(adapter, parent=sys)
        Or:
            vswfeed = VSwitch.get(adapter, parent_type=System,
                                  parent_uuid=sys_uuid)

        :param cls: A subclass of EntryWrapper.  Its schema_type will be used
                    as the first argument to adapter.read()
        :param adapter: The pypowervm.adapter.Adapter instance through which to
                        perform the GET.
        :param uuid: If retrieving a single entry, specify its string UUID.
                     For ROOT objects, you may specify either uuid or root_id;
                     for CHILD objects, you may specify either uuid or
                     child_id.
        :param parent_type: If the invoking class represents a CHILD, specify
                            either the parent parameter or BOTH parent_type and
                            parent_uuid. This parameter may be either the
                            schema_type or the EntryWrapper subclass of the
                            parent ROOT object.
        :param parent_uuid: If the invoking class represents a CHILD, specify
                            either the parent parameter or BOTH parent_type and
                            parent_uuid. This parameter indicates the UUID of
                            the parent ROOT object. Do not use the root_id
                            parameter.
        :param parent: If the invoking class represents a CHILD, specify either
                       the parent parameter or BOTH parent_type and
                       parent_uuid.  This parameter is an EntryWrapper
                       representing the parent ROOT object of the CHILD to be
                       retrieved.
        :param read_kwargs: Any arguments to be passed directly through to
                            Adapter.read().
        :return: An EntryWrapper (or list thereof) around the requested REST
                 object.  (Note that this may not be of the type from which the
                 method was invoked, e.g. if the child_type parameter is used.)
        """
        parent_type, parent_uuid = util.parent_spec(parent, parent_type,
                                                    parent_uuid)
        if parent_type is not None:
            # CHILD mode
            resp = cls._read_child(adapter, parent_type, parent_uuid, uuid,
                                   read_kwargs)
        else:
            # ROOT mode
            if any(k in read_kwargs for k in ('child_type', 'child_id')):
                raise ValueError(_("Developer error: specify 'parent' or "
                                   "('parent_type' and 'parent_uuid') to "
                                   "retrieve a CHILD object."))
            if uuid is not None:
                if 'root_id' in read_kwargs:
                    raise ValueError(_("Specify either 'uuid' or 'root_id' "
                                       "when requesting a ROOT object."))
                read_kwargs['root_id'] = uuid
            resp = adapter.read(cls.schema_type, **read_kwargs)
        return cls.wrap(resp)

    @classmethod
    def _read_child(cls, adapter, parent_type, parent_uuid, uuid, read_kwargs):
        """Helper method for 'get' to read CHILD feed or entry Response.

        Params are as described in the 'get' method.
        """
        if parent_uuid is None:
            raise ValueError(_("Both parent_type and parent_uuid are required "
                               "when retrieving a CHILD feed or entry."))
        if 'root_id' in read_kwargs:
            raise ValueError(_("Specify the parent's UUID via the parent_uuid "
                               "parameter."))
        if uuid is not None:
            if 'child_id' in read_kwargs:
                raise ValueError(_("Specify either 'uuid' or 'child_id' when "
                                   "requesting a CHILD object."))
            read_kwargs['child_id'] = uuid
        # Accept parent_type as either EntryWrapper subclass or string
        if not isinstance(parent_type, str):
            parent_type = parent_type.schema_type
        return adapter.read(parent_type, root_id=parent_uuid,
                            child_type=cls.schema_type, **read_kwargs)

    @classmethod
    def get_by_href(cls, adapter, href, **rbh_kwargs):
        """Get a wrapper or feed given a URI.

        This can be useful for retrieving wrappers "associated" with other
        wrappers, where the association is provided via an atom link.  Some
        examples are TrunkAdapter.associated_vswitch_uri and
        VNICBackDev.vios_href.

        :param adapter: A pypowervm.adapter.Adapter instance for REST API
                        communication.
        :param href: The string URI (including scheme://host:port/) of the
                     entry or feed to retrieve.
        :param rbh_kwargs: Keyword arguments to be passed directly to Adapter's
                           read_by_href method.
        :return: EntryWrapper subclass of the appropriate type, or a list
                 thereof, representing the entry/feed associated with the href
                 parameter.
        """
        return cls.wrap(adapter.read_by_href(href, **rbh_kwargs))

    @classmethod
    def search(cls, adapter, negate=False, xag=None, parent_type=None,
               parent_uuid=None, one_result=False, parent=None, **kwargs):
        """Performs a REST API search.

        Searches for object(s) of the type indicated by cls having (or not
        having) the key/value indicated by the (single) kwarg.

        Regular expressions, comparators, and logical operators are not
        supported.

        :param cls: A subclass of EntryWrapper.  The wrapper class may define
                    a search_keys member, which is a dictionary mapping a
                    @property getter method name to a search key supported by
                    the REST API for that object type.  To retrieve an XML
                    report of the supported search keys for object Foo,
                    perform: read('Foo', suffix_type='search').
                    If the wrapper class does not define a search_keys member,
                    OR if xag is None, the fallback search algorithm performs a
                    GET of the entire feed of the object type and loops through
                    it looking for (mis)matches on the @property indicated by
                    the search key.
        :param adapter: The pypowervm.adapter.Adapter instance through which to
                        perform the search.
        :param negate: If True, the search is negated - we find all objects of
                       the indicated type where the search key does *not* equal
                       the search value.
        :param xag: List of extended attribute group names.
        :param parent_type: If searching for CHILD objects, specify either
                            the parent parameter or BOTH parent_type and
                            parent_uuid.  This parameter indicates the parent
                            ROOT object.  It may be either the string schema
                            type or the corresponding EntryWrapper subclass.
        :param parent_uuid: If searching for CHILD objects, specify either
                            the parent parameter or BOTH parent_type and
                            parent_uuid.  This parameter specifies the UUID of
                            the parent ROOT object.  If parent_type is
                            specified, but parent_uuid is None, all parents of
                            the ROOT type will be searched. This may result in
                            a slow response time.
        :param one_result: Use when expecting (at most) one search result.  If
                           True, this method will return the first element of
                           the search result list, or None if the search
                           produced no results.
        :param parent: If searching for CHILD objects, specify either the
                       parent parameter or BOTH parent_type and parent_uuid.
                       This parameter is an EntryWrapper instance indicating
                       the parent ROOT object.
        :param kwargs: Exactly one key=value.  The key must correspond to a key
                       in cls.search_keys and/or the name of a getter @property
                       on the EntryWrapper subclass.  Due to limitations of
                       the REST API, if specifying xags or searching for a
                       CHILD, the key must be the name of a getter @property.
                       The value is the value to match.
        :return: If one_result=False (the default), a list of instances of the
                 cls.  The list may be empty (no results were found).  It may
                 contain more than one instance (e.g. for a negated search, or
                 for one where the key does not represent a unique property of
                 the object).  If one_result=True, returns a single instance of
                 cls, or None if the search produced no results.
        """
        def list_or_single(results, single):
            """Returns either the results list or its first entry.

            :param results: The list of results from the search.  May be empty.
                            Must not be None.
            :param single: If False, return results unchanged.  If True, return
                           only the first entry in the results list, or None if
                           results is empty.
            """
            if not single:
                return results
            return results[0] if results else None

        try:
            parent_type, parent_uuid = util.parent_spec(parent, parent_type,
                                                        parent_uuid)
        except ValueError:
            # Special case where we allow parent_type without parent_uuid.  The
            # reverse is caught by the check below.
            if parent_type is not None and type(parent_type) is not str:
                parent_type = parent_type.schema_type

        # parent_uuid makes no sense without parent_type
        if parent_type is None and parent_uuid is not None:
            raise ValueError(_('Parent UUID specified without parent type.'))

        if len(kwargs) != 1:
            raise ValueError(_('The search() method requires exactly one '
                               'key=value argument.'))

        key, val = kwargs.popitem()
        try:
            # search API does not support xag or CHILD
            if xag is not None or parent_type is not None:
                # Cheater's way to cause _search_by_feed to be invoked
                raise AttributeError()
            search_key = cls.search_keys[key]
        except (AttributeError, KeyError):
            # Fallback search by [GET feed] + loop
            return list_or_single(
                cls._search_by_feed(adapter, cls.schema_type, negate, key, val,
                                    xag, parent_type, parent_uuid), one_result)

        op = '!=' if negate else '=='
        quote = urllib.parse.quote if six.PY3 else urllib.quote
        search_parm = "(%s%s'%s')" % (search_key, op, quote(str(val), safe=''))
        # Let this throw HttpError if the caller got it wrong.
        # Note that this path will only be hit for ROOTs.
        return list_or_single(
            cls.wrap(cls._read_parent_or_child(
                adapter, cls.schema_type, parent_type, parent_uuid,
                suffix_type='search', suffix_parm=search_parm)),
            one_result)

    @classmethod
    def _search_by_feed(cls, adapter, target_type, negate, key, val, xag,
                        parent_type, parent_uuid):
        if not hasattr(cls, key):
            raise ValueError(_("Wrapper class %(class)s does not support "
                               "search key '%(key)s'.") %
                             {'class': cls.__name__, 'key': key})
        feedwrap = cls.wrap(cls._read_parent_or_child(adapter, target_type,
                                                      parent_type, parent_uuid,
                                                      xag=xag))
        retlist = []
        val = str(val)
        for entry in feedwrap:
            entval = str(getattr(entry, key, None))
            include = (entval != val) if negate else (entval == val)
            if include:
                retlist.append(entry)
        return retlist

    @staticmethod
    def _read_parent_or_child(adapter, target_type, parent_type, parent_uuid,
                              **kwargs):
        if parent_type is None:
            # ROOT feed search
            return adapter.read(target_type, **kwargs)
        if parent_uuid is not None:
            # CHILD of a specific ROOT
            return adapter.read(parent_type, root_id=parent_uuid,
                                child_type=target_type, **kwargs)
        # Search all ROOTs of the specified type.
        ret = None
        # Wishing there was a quick URI to get all UUIDs.
        # Let EntryWrapper.wrap figure out the wrapper type.  Whatever it
        # is, the uuid @property is available.
        for parent in EntryWrapper.wrap(adapter.read(parent_type)):
            resp = adapter.read(
                parent_type, root_id=parent.uuid, child_type=target_type,
                **kwargs)
            # This is a bit of a cheat.  Technically extending the feed of
            # a Response doesn't result in a legal Response (the rest of
            # the metadata won't accurately reflect the feed).  However
            # this is guaranteed only to be used immediately by wrap() to
            # extract the Entrys.
            if ret is None:
                ret = resp
            else:
                ret.feed.entries.extend(resp.feed.entries)
        return ret

    def create(self, parent_type=None, parent_uuid=None, timeout=-1,
               parent=None):
        """Performs an adapter.create (REST API PUT) with this wrapper.

        :param parent_type: If creating a CHILD, specify either the parent
                            parameter or BOTH parent_type and parent_uuid.
                            This parameter may be either the schema_type or the
                            EntryWrapper subclass of the parent ROOT object.
        :param parent_uuid: If creating a CHILD, specify either the parent
                            parameter or BOTH parent_type and parent_uuid.
                            This parameter indicates the UUID of the parent
                            ROOT object.
        :param timeout: (Optional) Integer number of seconds after which to
                        time out the PUT request.  -1, the default, causes the
                        request to use the timeout value configured on the
                        Session belonging to the Adapter.
        :param parent: If creating a CHILD, specify either the parent parameter
                       or BOTH parent_type and parent_uuid.  This parameter is
                       an EntryWrapper representing the parent ROOT object of
                       the CHILD to be created.
        :return: New EntryWrapper of the invoking class representing the PUT
                 response.
        """
        service = pc.SERVICE_BY_NS[self.schema_ns]
        parent_type, parent_uuid = util.parent_spec(parent, parent_type,
                                                    parent_uuid)
        if parent_type is None and parent_uuid is None:
            # ROOT
            resp = self.adapter.create(self, self.schema_type, service=service,
                                       timeout=timeout)
        else:
            # CHILD
            resp = self.adapter.create(
                self, parent_type, root_id=parent_uuid,
                child_type=self.schema_type, service=service, timeout=timeout)
        return self.wrap(resp)

    def delete(self):
        """Performs an adapter.delete (REST API DELETE) with this wrapper."""
        self.adapter.delete_by_href(self.href, etag=self.etag)

    # TODO(IBM): Remove deprecated xag parameter
    def update(self, xag='__DEPRECATED__', timeout=-1):
        """Performs adapter.update of this wrapper.

        :param xag: DEPRECATED - do not use.
        :param timeout: (Optional) Integer number of seconds after which to
                        time out the POST request.  -1, the default, causes the
                        request to use the timeout value configured on the
                        Session belonging to the Adapter.
        :return: The updated wrapper, per the response from the Adapter.update.
        """
        if xag != '__DEPRECATED__':
            import warnings
            warnings.warn(
                _("The 'xag' parameter to EntryWrapper.update is deprecated!  "
                  "At best, using it will result in a no-op.  At worst, it "
                  "will give you incurable etag mismatch errors."),
                DeprecationWarning)
        if timeout == -1:
            # Override default timeout to 60 minutes unless the Session is
            # configured for longer already.
            timeout = max(60 * 60, self.adapter.session.timeout)

        # adapter.update_by_path expects the path (e.g.
        # '/rest/api/uom/Object/UUID'), not the whole href.
        path = util.dice_href(self.href, include_fragment=False)
        return self.wrap(self.adapter.update_by_path(self, self.etag, path,
                                                     timeout=timeout))

    @property
    def element(self):
        return self.entry.element

    @property
    def etag(self):
        return self._etag

    @property
    def href(self):
        """Finds the reference to the entity.

        Assumes that the entity has a link element that references self.  If
        it does not, returns None.
        """
        return self.entry.self_link

    @property
    def related_href(self):
        """Returns the URI to be used for references in other elements.

        This will return a root URI (no extended attributes, no fragments).
        This should be used as needed to support entries/elements that have
        relationships to others.
        """
        temp_href = self.href
        return util.dice_href(temp_href, include_scheme_netloc=True,
                              include_query=False, include_fragment=False)

    @property
    def _type_and_uuid(self):
        """Return the type and uuid of this entry together in one string.

        This is useful for error messages, logging, etc.
        """
        entry_type = self.schema_type
        uuid = self.uuid

        if entry_type is None:
            entry_type = self.__class__.__name__

        if uuid is None:
            uuid = "UnknownUUID"

        return entry_type + ":" + uuid


class ElementWrapper(Wrapper):
    """Base wrapper for Elements."""
    # If it's an Element, it's *probably* a DETAIL.  (Redundant assignment,
    # but prefer to be explicit.)
    has_metadata = False

    @classmethod
    def _bld(cls, adapter, tag=None, has_metadata=None, ns=None, attrib=None):
        """Create a fresh ElementWrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param tag: XML tag for the ElementWrapper's root Element.
        :param has_metadata: If True, a basic <Metadata/> child is created
                             under the root element.
        :param ns: XML namespace for the contents.
        :param attrib: XML attributes for the root element.
        """
        ret = cls()
        ret.element = cls._bld_element(
            adapter, tag=tag, has_metadata=has_metadata, ns=ns, attrib=attrib)
        return ret

    @classmethod
    def wrap(cls, element, **kwargs):
        """Wrap an existing adapter.Element OR construct a fresh one.

        This method should usually be invoked from an ElementWrapper subclass
        decorated by Wrapper.pvm_type, and an instance of that subclass will be
        returned.

        If invoked directly from ElementWrapper, we attempt to detect whether
        an appropriate subclass exists based on the Element's tag.  If so, that
        subclass is used; otherwise a generic ElementWrapper is used.

        :param element: An existing adapter.Element to wrap.
        :param **kwargs: Arbitrary attributes to set on the new ElementWrapper.
        :returns: An ElementWrapper (subclass) instance containing the element.
        """
        wcls = (cls._class_for_element(element)
                if cls.schema_type is None
                else cls)
        wrap = wcls()
        wrap.element = element
        for key, val in kwargs.items():
            setattr(wrap, key, val)
        return wrap

    @property
    def _type_and_uuid(self):
        """Return the type of this element.

        This is useful for error messages, logging, etc.
        """
        entry_type = self.schema_type

        if entry_type is None:
            entry_type = self.__class__.__name__

        return entry_type

    def __eq__(self, other):
        """Tests equality."""
        return self.element == other.element

    def __hash__(self):
        """Hash value.

        Necessary to be overwritten because of the side effect in Python 3.x
        of overwriting the __eq__ method causing an object to be unhashable.
        """
        return super(ElementWrapper, self).__hash__()


class WrapperElemList(list):
    """The wrappers can create complex Lists (from a Group from the response).

    The lists that they wrap tend to be generated on each 'get' from the
    property.  This list allows for modification of the 'wrappers' that
    get returned, which update the backing elements.

    This is not a full implementation of a list.  Only the 'common use' methods
    are supported

    Functions that are provided:
     - Getting via index (ex. list[1])
     - Obtaining the length (ex. len(list))
     - Extending the list (ex. list.extend(other_list))
     - Appending to the list (ex. list.append(other_elem))
     - Removing from the list (ex. list.remove(other_elem))
    """

    def __init__(self, root_elem, child_class=None, indirect=None, **kwargs):
        """Creates a new list backed by an Element anchor and child type.

        :param root_elem: The container element.  Should be the backing
                          element, not a wrapper.
                          Ex. The element for 'SharedEthernetAdapters'.
        :param child_class: The child class (subclass of ElementWrapper).
                            This is optional.  If not specified, will wrap
                            all children elements.
        :param indirect: Name of schema layer to ignore between root_elem and
                         the target child_class.  This is for schema structures
                         such as:
                            <IOAdapters>
                                <IOAdapterChoice>
                                    <IOAdapter>...</IOAdapter>
                                </IOAdapterChoice>
                                <IOAdapterChoice>
                                    <SRIOVAdapter>...</SRIOVAdapter>
                                </IOAdapterChoice>
                                ...
                          </IOAdapters>
                         In this case, we want WrapperElemList to return
                            [IOAdapter, SRIOVAdapter, ...]
                         ...ignoring the intervening <IOAdapterChoice/>
                         layer, so we would set indirect='IOAdapterChoice'.
                         Note that we rely upon the intervening layer (in this
                         example, IOAdapterChoice) to contain nothing but the
                         target element type - not even <Metadata/>.
        :param kwargs: Optional additional named arguments that may be passed
                       into the wrapper on creation.
        """
        self.root_elem = root_elem
        if child_class is not None:
            self.child_class = child_class
        else:
            # Default to the ElementWrapper, which should resolve to the
            # appropriate class type.
            self.child_class = ElementWrapper
        self.indirect = indirect
        self.injects = kwargs

    def __find_elems(self):
        root_elems = self.root_elem.findall(
            self.indirect) if self.indirect else [self.root_elem]
        found = []
        for root_elem in root_elems:
            if (self.child_class is not None and
                    self.child_class is not ElementWrapper):
                found.extend(root_elem.findall(self.child_class.schema_type))
            else:
                found.extend(list(root_elem))
        return found

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            all_elems = self.__find_elems()
            all_elems = all_elems[idx.start:idx.stop:idx.step]
            return [self.child_class.wrap(x, **self.injects)
                    for x in all_elems]

        elem = self.__find_elems()[idx]
        return self.child_class.wrap(elem, **self.injects)

    def index(self, value):
        elems = self.__find_elems()
        return elems.index(value.element)

    def __getslice__(self, i, j):
        elems = self.__find_elems()
        return [self.child_class.wrap(x, **self.injects)
                for x in elems[i:j]]

    def __len__(self, *args, **kwargs):
        return len(self.__find_elems())

    def __iter__(self):
        elems = self.__find_elems()
        for elem in elems:
            yield self.child_class.wrap(elem, **self.injects)

    def __str__(self):
        return '[' + ', '.join([str(self.child_class.wrap(
            elem, **self.injects)) for elem in self.__find_elems()]) + ']'

    def __repr__(self):
        return '[' + ', '.join([repr(self.child_class.wrap(
            elem, **self.injects)) for elem in self.__find_elems()]) + ']'

    def __contains__(self, item):
        elems = self.__find_elems()
        return item.element in elems

    def extend(self, seq):
        for elem in seq:
            self.append(elem)

    def append(self, elem):
        self.root_elem.element.append(
            _indirect_child_elem(elem, self.indirect).element)

    def remove(self, elem):
        find_elem = _indirect_child_elem(elem, self.indirect)
        # Try this way first...if there is a value error, that means
        # that the identical element isn't here...need to try 'functionally
        # equivalent' -> slower...
        try:
            self.root_elem.remove(find_elem)
            return
        except ValueError:
            pass

        # Onto the slower path.  Get children and see if any are equivalent
        children = list(self.root_elem)
        equiv = util.find_equivalent(find_elem, children)
        if equiv is None:
            raise ValueError(_('No such child element.'))
        self.root_elem.remove(equiv)


class ActionableList(list):
    """Provides a List that will call back to a function on modification.

    Does not support lower level modifications (ex. list[5] = other_elem),
    but does support extend, append, remove, insert and pop.
    """

    def __init__(self, list_data, action):
        """Creations the action list.

        :param list_data: The list data
        :param action: The action to call back to.  Should take in a list
                       as a parameter (this is then list post modification).
        """
        super(ActionableList, self).__init__(list_data)
        self.action = action

    def extend(self, seq):
        super(ActionableList, self).extend(seq)
        self.action(self)

    def append(self, elem):
        super(ActionableList, self).append(elem)
        self.action(self)

    def remove(self, elem):
        super(ActionableList, self).remove(elem)
        self.action(self)

    def insert(self, index, obj):
        super(ActionableList, self).insert(index, obj)
        self.action(self)

    def pop(self, index=-1):
        elem = super(ActionableList, self).pop(index)
        self.action(self)
        return elem


@six.add_metaclass(abc.ABCMeta)
class WrapperSetUUIDMixin(object):
    """Abstract mixin to enable a Wrapper instance to set its UUID.

    USE WITH CAUTION.  Caveats:
    This will only work on Wrappers with has_metadata=True.
    Not all elements accept a consumer-set UUID.  Of those that do, some
    only accept it at creation, not on update.
    """

    def set_uuid(self, new_uuid):
        """Set the UUID of the XML entity represented by this Wrapper.

        :param new_uuid: The UUID to set.  Must valid uuid type or string
                         properly formatted (e.g. 8-4-4-4-12)
        """
        if not self.has_metadata:
            raise AttributeError(
                _('Cannot set UUID on Wrapper with no Metadata.'))

        # Step 1: sanitize uuid value
        s_uuid = str(new_uuid)
        if s_uuid != pvm_uuid.convert_uuid_to_pvm(s_uuid):
            raise ValueError(_('uuid value not valid: %s') % new_uuid)

        # Step 2: (vivify and) set Metadata/Atom/AtomID
        atom = self._find('Metadata/Atom')
        atomid = atom.find('AtomID')
        if atomid is None:
            atomid = ent.Element('AtomID', self.adapter)
            atom.append(atomid)
        atomid.text = s_uuid

        # Step 3: if an Atom, update the properties['id'] to match
        try:
            self.entry.properties['id'] = s_uuid
        except AttributeError:
            # No entry (this is an ElementWrapper) - nothing to do.
            # Note: we don't trap KeyError: if entry.properties is there, but
            # doesn't have an 'id' key, something is wrong.
            pass


class EntryWrapperGetter(object):
    """Attribute container with enough information to GET an EntryWrapper.

    An instance of this class can be used to defer the REST call which fetches
    a PowerVM object.  This will typically be used to initialize a
    pypowervm.utils.transaction.WrapperTask, or as the first parameter to a
    method decorated as pypowervm.utils.transaction.entry_transaction, allowing
    that method to acquire a lock before performing the GET, thus minimizing
    the probability of out-of-band changes resulting in etag mismatch and
    requiring a retry.
    """
    def __init__(self, adapter, entry_class, entry_uuid, parent_class=None,
                 parent_uuid=None, xag=None, parent=None):
        """Create a GET specification for an EntryWrapper.

        :param adapter: A pypowervm.adapter.Adapter instance through which the
                        GET can be performed.
        :param entry_class: An EntryWrapper subclass indicating the type of the
                            entry to GET.
        :param entry_uuid: The string UUID of the entry to GET.
        :param parent_class: If the target object type is CHILD, specify either
                             the parent parameter or BOTH parent_class and
                             parent_uuid.  This param is the EntryWrapper
                             subclass of the ROOT parent object type.
        :param parent_uuid: If the target object type is CHILD, specify either
                            the parent parameter or BOTH parent_class and
                            parent_uuid.this param is the UUID of the ROOT
                            parent object.
        :param xag: List of extended attribute groups to request on the object.
        :param parent: If the target object type is CHILD, specify either the
                       parent parameter or BOTH parent_class and parent_uuid.
                       This parameter represents the ROOT parent object.
        """
        def validate_wrapper_type(var):
            if not issubclass(type(var), type) or not issubclass(var, Wrapper):
                raise ValueError(_("Must specify a Wrapper subclass."))
        self.adapter = adapter
        validate_wrapper_type(entry_class)
        self.entry_class = entry_class
        self.entry_uuid = entry_uuid
        parent_class, parent_uuid = util.parent_spec(parent, parent_class,
                                                     parent_uuid)
        if (parent_class and not parent_uuid) or (
                parent_uuid and not parent_class):
            raise ValueError(_("Must specify both parent class and parent "
                               "UUID, or neither."))
        self.parent_class = parent_class
        self.parent_uuid = parent_uuid
        self.xag = xag
        self.cache = None

    def get(self, refresh=False):
        """Return the EntryWrapper indicated by this instance.

        If the EntryWrapper has not yet been retrieved, it is fetched via GET
        from the REST API.  Thereafter, it is cached.  Subsequent calls to this
        method will return the cached copy unless refresh=True, in which case
        the cached copy is refreshed before returning.

        :param refresh: (Optional) If True, and the specified EntryWrapper was
                        previously retrieved, it is refreshed before being
                        returned.  If False (the default), it is returned
                        without refreshing.  If the specified EntryWrapper had
                        not yet been retrieved, this parameter has no effect.
        :return: The EntryWrapper specified by this EntryWrapperGetter
                 instance.
        """
        if self.cache is None:
            if self.parent_class:
                root_type = self.parent_class
                root_id = self.parent_uuid
                child_type = self.entry_class.schema_type
                child_id = self.entry_uuid
            else:
                root_type = self.entry_class.schema_type
                root_id = self.entry_uuid
                child_type = None
                child_id = None
            self.cache = self.entry_class.wrap(self.adapter.read(
                root_type, root_id, child_type=child_type, child_id=child_id,
                xag=self.xag))
        elif refresh:
            self.cache = self.cache.refresh()
        return self.cache

    @property
    def uuid(self):
        """Return the UUID of the entry for which this spec was created.

        This mainly exists so we can ask for wrapper_or_spec.uuid.
        """
        return self.entry_uuid


class FeedGetter(EntryWrapperGetter):
    """Attribute container with enough information to GET an EntryWrapper feed.

    An instance of this class can be used to defer the REST call which fetches
    a feed of PowerVM objects (a list of EntryWrapper).  This will typically be
    used to initialize a pypowervm.utils.transaction.FeedTask, allowing the
    FeedTask to defer the GET as long as possible, thus minimizing the
    probability of out-of-band changes resulting in etag mismatch and requiring
    a retry.
    """
    def __init__(self, adapter, entry_class, parent_class=None,
                 parent_uuid=None, xag=None, parent=None):
        """Create a GET specification for an EntryWrapper feed.

        :param adapter: A pypowervm.adapter.Adapter instance through which the
                        GET can be performed.
        :param entry_class: An EntryWrapper subclass indicating the type of the
                            feed to GET.
        :param parent_class: If the target object type is CHILD, specify either
                             the parent parameter or BOTH parent_class and
                             parent_uuid.  This param is the EntryWrapper
                             subclass of the ROOT parent object type.
        :param parent_uuid: If the target object type is CHILD, specify either
                            the parent parameter or BOTH parent_class and
                            parent_uuid.this param is the UUID of the ROOT
                            parent object.
        :param xag: List of extended attribute groups to request on the feed.
        :param parent: If the target object type is CHILD, specify either the
                       parent parameter or BOTH parent_class and parent_uuid.
                       This parameter represents the ROOT parent object.
        """
        # Using entry_uuid=None will cause the GET to fetch the feed.
        super(FeedGetter, self).__init__(
            adapter, entry_class, None, parent=parent,
            parent_class=parent_class, parent_uuid=parent_uuid, xag=xag)

    def get(self, refresh=False, refetch=False):
        """Return the feed (list of EntryWrappers) indicated by this instance.

        If the feed has not yet been retrieved, it is fetched via GET from the
        REST API.  Thereafter, it is cached.  Subsequent calls to this
        method will return the cached copy unless refresh or refetch is
        specified.

        The refresh option, if True, will cause each entry in the feed to be
        refreshed if previously cached.  The refetch option, if True, will
        cause the feed to be refetched as a whole.

        Note: due to the design of the REST server, refetch will generally
        perform better than refresh.

        :param refresh: (Optional) If True, and the specified feed was
                        previously retrieved, each entry therein is refreshed
                        before the feed is returned.  If the specified feed had
                        not yet been retrieved, this parameter has no effect.
                        If both refresh and refetch are True, refresh takes
                        precedence.
        :param refetch: (Optional) If True, a fresh GET of the entire feed is
                        performed, regardless of whether the feed was fetched
                        and cached previously.
                        If both refresh and refetch are True, refresh takes
                        precedence.
        :return: The feed (list of EntryWrappers) specified by this FeedGetter
                 instance.
        """
        # Note: self.cache is the feed (list of EntryWrapper) in the context of
        # this subclass.  Therefore, the superclass's concept of 'refresh' is
        # no good (it would be trying [ewrap, ...].refresh()).
        if refresh and self.cache is not None:
            # Future: parallelize, for what it's worth.
            new_feed = [ewrap.refresh() for ewrap in self.cache]
            self.cache = new_feed
            return self.cache

        # To refetch, simply wipe the cache before super.get().
        if refetch:
            self.cache = None

        # Never, never call super.get(refresh=True).
        return super(FeedGetter, self).get(refresh=False)


class UUIDFeedGetter(FeedGetter):
    """Quasi-FeedGetter that builds its "feed" based on a list of UUIDs.

    This is expected to be useful for building FeedTasks when, for example:
    - The FeedTask is operating on an SSP (the VIOSes aren't necessarily all in
      the same feed);
    - The operation is only concerned with one REST object, but a WrapperTask
      is not sufficient.
    """
    def __init__(self, adapter, entry_class, uuid_list, parent_class=None,
                 parent_uuid=None, xag=None, parent=None):
        """Create a UUIDFeedGetter.

        :param adapter: See FeedGetter.
        :param entry_class: See FeedGetter.
        :param uuid_list: Iterable of string UUIDs of the objects with which to
                          populate the quasi-feed.
        :param parent_class: See FeedGetter.
        :param parent_uuid: See FeedGetter.
        :param xag: See FeedGetter.
        :param parent: See FeedGetter.
        """
        super(UUIDFeedGetter, self).__init__(
            adapter, entry_class, parent=parent, parent_class=parent_class,
            parent_uuid=parent_uuid, xag=xag)
        self.uuid_list = uuid_list
        self._create_wrapper_getters()

    def _create_wrapper_getters(self):
        self.wrapper_getters = [EntryWrapperGetter(
            self.adapter, self.entry_class, entry_uuid,
            parent_class=self.parent_class, parent_uuid=self.parent_uuid,
            xag=self.xag) for entry_uuid in self.uuid_list]

    def get(self, refresh=False, refetch=False):
        """Get the individual wrappers for each UUID and put them in a 'feed'.

        :param refresh: See FeedGetter.get.
        :param refetch: See FeedGetter.get.
        """
        if refetch:
            # Rebuild the wrapper getters, guaranteeing that we clear anything
            # already fetched
            self._create_wrapper_getters()
        # Populate the quasi-feed from the individual wrapper getters.
        # Future: parallelize, for what it's worth.
        return [wg.get(refresh=refresh) for wg in self.wrapper_getters]
